void get_peelseq (int **psq)
{
   /* Note: cmplist is a one-indexed global given in nghds.h          */
   /* Note: The first node of the Indlist linked list is NULL,        */
   /*       whereas, the last node of the Marlist linked list is NULL */
   /* Note: nghd will generally not have the same ordering            */
   /*       as the input pedigree file however, nghd does order       */
   /*       each member of each component consecutively               */
   /* Note: MyIndPtr is zero-indexed as is nghds, and the structure   */
   /*       definition is in Autozyg/Utils/map.h                      */
   /* Note: The term nome denotes a non-final with multiple           */
   /*       connections to other members of the pedigree. In          */
   /*       particular, any individual with multiple marriages is     */
   /*       nome as is any individual with both ancestors and         */
   /*       descendants.                                              */
   /* Note: Founder status, peeled status, marital status, number     */
   /*       of spouses, nome status and previous R-function status    */
   /*       are all changed dynamically.                              */
   /* Note: The program concludes by peeling down onto a child whose  */
   /*       previous R-functs have all been taken into account.       */

   int   c, i, j, k, mom, dad, famsz, unit, notstuck, done;
   int   pivot, onlychild, mtr, ftr, m, a, b, x;
   Indlist *Who;
   Marlist *Spouse;

   a = 0;
   b = 0;
   x = 0;

   MyIndPtr = malloc (pedsize * sizeof (MyInd));
   for (i = 0; i < pedsize; i++)
   {
      /* Initialize all of the status indicators */
      MyIndPtr[i].prev_rf = MyIndPtr[i].peeled = 0;
      MyIndPtr[i].married = MyIndPtr[i].nspouses = 0;
      MyIndPtr[i].foun = MyIndPtr[i].nome = 0;

      Spouse = nghd[i].marriages;
      if (Spouse != NULL)
         MyIndPtr[i].married = 1;
      while (Spouse != NULL)
      {
         MyIndPtr[i].nspouses++;
         Spouse = Spouse->next;
      }
      if (MyIndPtr[i].nspouses > 1)
         MyIndPtr[i].nome = 1;
      if ((nghd[i].pa)->index != -1 && MyIndPtr[i].married)
         MyIndPtr[i].nome = 1;
      if ((nghd[i].pa)->index == -1)
         MyIndPtr[i].foun = 1;

      /* Use this function to see the information on indiv. i */
      /* pr_person(i); */
   }

   for (c = 1; c <= nCompnt; c++)       /* for each family    */
   {
      for (j = 0; j <= MXPSQ; j++)
         psq[c][j] = -1;
      notstuck = 1;
      done = 0;
      famsz = cmplist[c].size;
      while (notstuck == 1 && done == 0) /* keep processing family until done */
      {
         notstuck = 0;
         k = 1;
         Who = cmplist[c].people;       /* first node of Indlist is always NULL */
         while (k <= famsz)             /* cycle thru all members until stuck or done */
         {
            Who = Who->next;
            unit = (Who->current)->index; /* unit is some family member's id */
            mom = ((Who->current)->ma)->index; /* unit's mom */
            dad = ((Who->current)->pa)->index; /* unit's dad */
            onlychild = 0;
            if (mom != -1 && dad != -1) /* is unit a non-founder */
            {
               for (m = 0; m < pedsize; m++)
               {
                  /* this search routine is not efficient,   */
                  /* but it only needs to be done at setup.  */
                  mtr = (nghd[m].ma)->index;
                  ftr = (nghd[m].pa)->index;
                  if (mtr == mom && ftr == dad)
                     onlychild++;
               }
               if (x == 0 && onlychild == 1 && MyIndPtr[unit].peeled == 0)
                  b = 0;
            }
            a = MyIndPtr[unit].married + MyIndPtr[unit].peeled;

            /* identify a potentially peelable cluster: if unit is a final, or if */
            /* unit is the only child of parents for whom all OTHER upper and     */
            /* lower R-functs have been considered, then unit's cluster may be    */
            /* peelable. Note that in each case, unit is always a child relative  */
            /* the potentially pealable cluster that is being considered.         */
            if (a == 0 || b == 0)
            {
               x = MyIndPtr[mom].nome + MyIndPtr[dad].nome;
               if (MyIndPtr[mom].peeled == 1 || MyIndPtr[dad].peeled == 1)
                  x = 2;
               if (a == 0 && x == 1)
               {                        /* one of the parents is potentially a pivot */
                  pivot = (MyIndPtr[mom].nome == 1) ? mom : dad;
                  notstuck = get_pivot (mom, dad, &pivot, c, psq, 1);
               }
               if (x == 0)
               {
                  pivot = nghd[unit].index; /* one of the kids is potential pivot */
                  notstuck = get_pivot (mom, dad, &pivot, c, psq, 0);
                  if (notstuck == 1 && MyIndPtr[pivot].peeled == 1)
                     done = 1;
               }
            }
            k++;
         }
         if (!notstuck)
         {
            j = 1;
            printf ("-1 ");             /* if fam complex --> length of psq unkonwn */
            while (psq[c][j] != -1)
            {
               printf (" %d ", psq[c][j]);
               j++;
            }
            printf ("\n\n Family %d is not zero-looped. Program quitting. \n",
                    c);
            exit (1);
         }
      }

      /* This code prints out the complete psq for zero-looped families with  */
      /* ids given according to the nghd indexing. If ids relative to the     */
      /* input pedigree names are needed, use %s and nghd[psq[c][j]].self     */
      /* for(j=0; j<=psq[c][0]; j++) printf(" %d ",psq[c][j]); printf("\n");  */
      /* printf("\n %d ", psq[c][0]);                                         */
      /* for(j = 1; j<=psq[c][0]; j++) printf(" %s ",nghd[psq[c][j]].self);   */
      /* printf("\n\n");                                                      */
   }
   free (MyIndPtr);
}

int get_pivot (int mother, int father, int *pvt, int fam, int **psq, int pflag)
{
   int   i, k = 0, t = 0, count = 0, mtr, ftr, kids = 0;
   int   success = 0, pnpvt, ppvt_flag = 0;
   static int j;

   ppvt_flag = pflag;
   for (i = 0; i < pedsize; i++)
   {
      mtr = (nghd[i].ma)->index;
      ftr = (nghd[i].pa)->index;
      if (mtr == mother && ftr == father) /* find siblings in this cluster */
      {
         kids++;
         count += MyIndPtr[i].married;
         if (!ppvt_flag && MyIndPtr[i].married == 1)
            *pvt = nghd[i].index;
      }
   }

   if (ppvt_flag == 1 && count == 0)    /* CLUSTER HAS A PARENT PIVOT    */
   {
      success = 1;
      psq[fam][1 + j] = 1;
      psq[fam][2 + j] = *pvt;
      if (mother != *pvt)
      {
         MyIndPtr[mother].peeled = 1;
         pnpvt = mother;
      }
      else
      {
         MyIndPtr[father].peeled = 1;
         pnpvt = father;
      }
      MyIndPtr[*pvt].nspouses--;

      if (MyIndPtr[*pvt].nspouses == 0)
         MyIndPtr[*pvt].married = MyIndPtr[*pvt].nome = 0;

      if (MyIndPtr[*pvt].foun == 1 && MyIndPtr[*pvt].nspouses == 1)
         MyIndPtr[*pvt].nome = 0;

      psq[fam][3 + j] = 1;
      psq[fam][5 + j] = 2;
      psq[fam][6 + j] = mother;
      psq[fam][7 + j] = father;
      psq[fam][8 + j] = kids;
      for (i = 0; i < pedsize; i++)
      {
         mtr = (nghd[i].ma)->index;
         ftr = (nghd[i].pa)->index;
         if (mtr == mother && ftr == father) /* find each child of this cluster */
         {
            k++;
            psq[fam][8 + j + k] = nghd[i].index;
            MyIndPtr[i].peeled = 1;
            if (MyIndPtr[i].prev_rf)    /* if kids have any  prev_rfs, get them */
            {
               t++;
               psq[fam][(8 + kids + j + t)] = 1;
               t++;
               psq[fam][(8 + kids + j + t)] = nghd[i].index;
               MyIndPtr[i].prev_rf = 0;
            }
         }
      }
      if (MyIndPtr[pnpvt].prev_rf)      /* if parent non pivot has prev_rf, get it  */
      {
         t++;
         psq[fam][(8 + kids + j + t)] = 1;
         t++;
         psq[fam][(8 + kids + j + t)] = nghd[pnpvt].index;
         MyIndPtr[pnpvt].prev_rf = 0;
      }
      if (MyIndPtr[*pvt].prev_rf)       /* if parent pivot has a prev_rf, get it  */
      {
         t++;
         psq[fam][(8 + kids + j + t)] = 1;
         t++;
         psq[fam][(8 + kids + j + t)] = nghd[*pvt].index;
      }
      MyIndPtr[*pvt].prev_rf = 1;       /* record prev_rf for this parent peel */
      psq[fam][4 + j] = t / 2;

      j = 8 + j + kids + t;             /* j holds the end position for ea. cluster */
   }

   if (ppvt_flag == 0 && count <= 1)    /*   CLUSTER HAS A KID PIVOT     */
   {
      success = 1;
      if (count)
         MyIndPtr[*pvt].foun = 1;
      if (MyIndPtr[*pvt].nspouses == 1)
         MyIndPtr[*pvt].nome = 0;
      psq[fam][1 + j] = 1;
      psq[fam][2 + j] = *pvt;
      MyIndPtr[mother].peeled = MyIndPtr[father].peeled = 1;
      MyIndPtr[mother].nspouses--;
      MyIndPtr[father].nspouses--;
      psq[fam][3 + j] = 1;
      psq[fam][5 + j] = 2;
      psq[fam][6 + j] = mother;
      psq[fam][7 + j] = father;
      psq[fam][8 + j] = kids;
      for (i = 0; i < pedsize; i++)
      {
         mtr = (nghd[i].ma)->index;
         ftr = (nghd[i].pa)->index;
         if (mtr == mother && ftr == father) /* find each kid in cluster */
         {
            k++;
            psq[fam][8 + j + k] = nghd[i].index;
            if (i != *pvt)
               MyIndPtr[i].peeled = 1;
            if (MyIndPtr[i].prev_rf)    /* if prev_rfs exist, get them */
            {
               t++;
               psq[fam][(8 + kids + j + t)] = 1;
               t++;
               psq[fam][(8 + kids + j + t)] = nghd[i].index;
               MyIndPtr[i].prev_rf = 0;
            }
         }
      }
      if (MyIndPtr[mother].prev_rf)     /* if mom has a prev_rf, get it  */
      {
         t++;
         psq[fam][(8 + kids + j + t)] = 1;
         t++;
         psq[fam][(8 + kids + j + t)] = nghd[mother].index;
         MyIndPtr[mother].prev_rf = 0;
      }
      if (MyIndPtr[father].prev_rf)     /* if dad has a prev_rf, get it  */
      {
         t++;
         psq[fam][(8 + kids + j + t)] = 1;
         t++;
         psq[fam][(8 + kids + j + t)] = nghd[father].index;
         MyIndPtr[father].prev_rf = 0;
      }
      MyIndPtr[*pvt].prev_rf = 1;       /* record prev_rf for this child peel */
      psq[fam][4 + j] = t / 2;

      j = 8 + j + kids + t;             /* j holds the end position for ea. cluster   */

      if (count == 0)                   /*  this is the final cluster  */
      {
         MyIndPtr[*pvt].peeled = 1;
         psq[fam][1 + j] = psq[fam][2 + j] = 0;
         psq[fam][3 + j] = 1;
         psq[fam][4 + j] = 2;
         psq[fam][5 + j] = psq[fam][6 + j] = 0;
         psq[fam][0] = j + 6;
         j = 0;
      }
   }
   return (success);
}

void pr_person (int j)                  /* prints out the contents of MyInd */
{
   MyInd x;
   x = MyIndPtr[j];
   printf ("Nghd id: %d  Ped id: %s \n", j, nghd[j].self);
   printf ("  -previous r-functions: %d \n", x.prev_rf);
   printf ("  -peeled status:        %d \n", x.peeled);
   printf ("  -married status:       %d \n", x.married);
   printf ("  -number of marriages:  %d \n", x.nspouses);
   printf ("  -founder status:       %d \n", x.foun);
   printf ("  -nome status:          %d \n", x.nome);
}

void prev_functions (int n1, int n2, int *seq, int *ptr)
{
   int   i, j, k, l;
   for (i = n1; i <= n2; ++i)
   {
      nprlik[i] = seq[ptr[0]];
      ptr[0]++;

      for (j = 1; j <= nprlik[i]; ++j)
      {
         iprlik[j][i] = seq[ptr[0]];
         ptr[0]++;
      }
      if (nprlik[i] == 0)
      {
         j = nstore;
         nprlik[i] = nindi[nstore];
         k = nprlik[i] + 1;
      }

      else
      {
         for (j = 2; j <= nstore; ++j)
         {
            if ((nindi[j] == nprlik[i]) && (ndex[j] != 0))
            {
               k = 1;
               while ((k <= nprlik[i]) && (iprlik[k][i] == namsto[j][k]))
                  ++k;
               if (k == nprlik[i] + 1)
                  break;
            }
         }
      }

      if (j < nstore + 1)               /* prev lhd i found at stored lhd j */
      {
         nstlik[i] = nstart[j];
         ndex[j] = 0;
         maxlik[i] = 0;
         for (l = 1; l <= nprlik[i]; ++l)
         {
            k = 1;
            while ((k <= ninvol) && (namsto[j][l] != invol[k]))
               ++k;
            if (k == ninvol + 1)
            {
               ++ninvol;
               invol[ninvol] = namsto[j][l];
            }
            iprlik[l][i] = k;
            if (k > maxlik[i])
               maxlik[i] = k;
         }
      }

      else                              /* prev lhd i not found in store */
      {
         if (nprlik[i] != 1)
         {
            printf ("\n Exited in prev lhd\n");
            exit (1);
         }
         l = iprlik[1][i];
         nstlik[i] = 0;
         k = 1;
         while ((k <= ninvol) && (l != invol[k]))
            ++k;
         if (k == ninvol + 1)
            ++ninvol;
         iprlik[1][i] = k;
         maxlik[i] = k;
      }
   }
}

void repack ()
{
   int   ntrans, newst;
   int   i, j, nb, nt;
   ntrans = 0;
   newst = 0;
   for (i = 1; i <= nstore; ++i)
   {
      if (ndex[i] == 0)
         ntrans += nstart[i + 1] - nstart[i];
      else
      {
         newst += 1;
         if (ntrans > 0)
         {
            nt = nstart[i + 1] - nstart[i];
            nb = nstart[i];
            for (j = 1; j <= nt; ++j)
               store[nb + j - ntrans] = store[nb + j];
            ndex[newst] = 1;
            nstart[newst + 1] = nb + nt - ntrans;
            nindi[newst] = nindi[i];
            for (j = 1; j <= nindi[i]; ++j)
               namsto[newst][j] = namsto[i][j];
         }
      }
   }
   nstore = newst;
}

int rescale (ff, nn)
     double *ff;
     int nn;
{
   int   j;
   double fm;

   fm = 0.0;
   for (j = 1; j <= nn; ++j)
      if (ff[j] > fm)
         fm = ff[j];

   if (fm == 0.0)
   {
      printf ("\n Exited in rescale\n");
      exit (1);
   }
   if (fm <= SMALL)
   {
      for (j = 1; j <= nn; ++j)
         ff[j] *= SCALE;
      return (1);
   }
   return (0);
}

void varlik (ff, nn)
     double *ff;
     int nn;
{
   int   nzero, j;
   double fm, fv;
   nzero = 0;
   fm = 0.0;
   fv = 0.0;
   for (j = 1; j <= nn; ++j)
   {
      if (ff[j] == 0.0)
         ++nzero;
      else
      {
         fm += ff[j];
         fv += ff[j] * ff[j];
      }
   }
   fm = fm / nn;
   fv = (fv / nn) - fm * fm;
}
