/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*       genedrop.h:  header file for genedrop.c                       */
/*                    contains #includes for header files for          */
/*                        libraries used                               */
/*                    and #defines, typedefs, variables and            */
/*                        function prototypes for file genedrop.c      */
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef GENEDROP_H
#define GENEDROP_H

/**
 * @defgroup GDrop Genedrop programs.
 * @ingroup Programs
 * @{
 */

#define DEFINE_DATA

#include "parsie.h"
#include "pedchk.h"

#ifdef IBD
#include "lsample.h"
#include "msample.h"
#include "markers.h"
#include "traits.h"     /* needed for trait structure def for prnt_ibdprobs.c */
#endif /*IBD*/

#ifdef MDP
#include "traits.h"
#include "quantset.h"
#endif /* MDP */
#include "rans.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif /* DMALLOC */

#define GM 0                    /* grandmother                          */
#define GF 1                    /* grandfather                          */
#define FOUNDER -1              /* founder                              */

FILE * pedF;
FILE * oPedF;
FILE * oSeedF;

                        /*kDat:  maternal & paternal inheritance        */
struct kDat {
    int kMat;                   /*maternal grandma's or grandpa's allele*/
    int kPat;                   /*paternal grandma's or grandpa's allele*/
            };                  /* ('GM' or 'GP')                       */
typedef struct kDat KDAT;
typedef KDAT * KDAT_PTR;
KDAT_PTR pKDat;

                        /*lablDat:  label for each locus                */
struct lablDat {
    int lablMa;                         /*label inherited from mother   */
    int lablPa;                         /*label inherited from father   */
               };
typedef struct lablDat LABL;
typedef LABL * LABL_PTR;
LABL_PTR pLablDat;

                        /*genoDat:  genotype for each locus             */
struct genoDat {
    int genoMa;                         /*allele inherited from mother  */
    int genoPa;                         /*allele inherited from father  */
               };
typedef struct genoDat GENO;
typedef GENO * GENO_PTR;
GENO_PTR pGenoDat;

                        /*phenoDat: phenotype for each trait locus      */
struct phenoDat {
    double addi;                        /*additive component            */
    double resi;                        /*residual component            */
    double quant;                       /*quantitative trait value      */
                };
typedef struct phenoDat PHENO;
typedef PHENO * PHENO_PTR;
PHENO_PTR pPhenoDat;

static struct _cleanup clean_drop[] = {
{"di",    free_fnams},
{"di",    freeChkPed},
{"di",    freeKey},
{"d ",    free_drop},
#ifdef IBD
{"i ",    free_ped},
{"i ",    free_cmplists},
{"i ",    free_nghds},
{"i ",    free_xxgg},
{"i ",    free_switches},
{"i ",    free_dub},
#endif /*IBD*/
{""  ,    NULL}} ;

void     allocGenes(void),      labelFounderGenes(void),
         dropMarkerGenes(void), dropTraitGenes(void),
         openSimPed(void),      writSimPed(void),
         openSeedOut(void),     freeStuff(void);

void     foundGenes(ALLELE_PTR, KDAT_PTR, GENO_PTR, int),
         inheritGenes(KEYS_PP, RMAP_PTR, KDAT_PTR, GENO_PTR, int),
         inheritLabels(KEYS_PP, KDAT_PTR, LABL_PTR, int),
         foundTrait(PHENO_PTR, TPARS_PTR),
         inheritTrait(KEYS_PP, PHENO_PTR, TPARS_PTR, int),
         makQuant(GENO_PTR, PHENO_PTR, TPARS_PTR);

/** @} */

#endif /*GENEDROP_H*/

