/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*
 *       lsample.h:  header file for the Lsample/ library                     *
 *                -  included by Lsample/ (locus sampling) functions          *
 *                   and by main programs calling Lsample/ functions          *
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef LSAMPLE_H
#define LSAMPLE_H

/**
 * @defgroup sample libsample
 * @ingroup Libraries
 * @{
 */

#include "incs.h"
#include "nghds.h"
#include "markers.h"
#include "rans.h"
#include "lk_long.h"

#define ALLOC_FAIL give_up_where(WHERE,ERS_mal,strerror(errno))

#define WAS_PIVOT 16
#define PEELED 32
#define JNT_PIVOT 64
#define STABLE_FLAGS 127
#define HAP_M 128
#define HAP_P 256
#define HAP_MAT 512
#define HAP_PAT 1024
#define HAP_JNT 2048
#define IN_RF 4096
#define HAD_M 8192
#define HAD_P 16384

#define X_MAT 0
#define X_PAT 1

#define X_MM_PM 0
#define X_MM_PP 1
#define X_MP_PM 2
#define X_MP_PP 3

/* Peeling output level - controlled by lower 3 bits in peel_trace */
#define TRACE_LEVEL_0 0
#define TRACE_LEVEL_1 1
#define TRACE_LEVEL_2 2
#define TRACE_LEVEL_3 3
#define TRACE_LEVEL_4 4
#define TRACE_MASK 7

#define TRACE_DUMP_SEQUENCE 16              /* Print out peeling sequence */
#define TRACE_NO_RECODE 32                  /* Don't recode alleles within
                                               pedigrees */
#define TRACE_NO_SETRECODE 64               /* Don't perform set recoding */
#define TRACE_NO_FAMILYPEEL 128             /* Don't perform any family based
                                               peeling */
#define TRACE_NO_PARTIALPEEL 256            /* Don't try to partially peel
                                               families */
#define USE_METIS 1024                      /* Use METIS library to get
                                               peeling sequence
                                               rather than amdbar */

#define CHK_PEEL(x) ((peel_trace&TRACE_MASK)>=(x))
#define PEEL_SIMPLE 1
#define PEEL_COMPLEX 2

#define SAMPLED_MAT 128
#define SAMPLED_PAT 256

#define OP_SAMPLING 128

#define ABT_FUNC(msg) abt(__FILE__,__LINE__,"%s(): %s",FUNC_NAME,msg)

#define HASHTABLE_SIZE 2053

struct fset
{
   double p;
   char  pat_gene[2];
   char  mat_gene[2];
};

#define TRAIT_LOCUS 1
#define MARKER_LOCUS 2

struct Locus
{
   int   locus_type;
   int   index;
   int   data_index;
   int   n_all, *n_alls;            /* No. alleles (overall and in each
                                       component) */
   double *g_mu;                    /* Genotype means  or penetrances */
   double va,ve;                    /* additive and residual variances      */
   int  *gt[2];                     /* Sampled genotypes */
   int  *seg[2];                    /* Sampled SI's */
   int  *genes[2];                  /* Founder gene array (current version) */
   int  *genes1[2];                 /* Founder gene array (alternate being
                                       tested) */
   int   max_alternates;
   int   chromosome;
   struct R_Func **r_func;
   int  *n_rfuncs, sparse_flag;
   int **allele_trans;
   struct Peelseq_Head *peel;
   double malepos, fempos, *freq;
   unsigned long **all_set, *req_set[3];
   void  (*penetrance) (double *, const Ind *, const struct Locus *, double);
};

struct peel_mem_block
{
   struct peel_mem_block *next;
   lk_ulong *index;
   double *val;
   size_t size, ptr;
};

union Peelseq_Pointer
{
   struct Complex_Element *complex;
   struct Simple_Element *simple;
};

struct Family
{
   Ind  *sire, *dam;
   Indlist *off;
   int   nkids;
   int   flag;
};

struct Peelseq_Head
{
   union Peelseq_Pointer ptr;
   int   type;
};

struct Simple_Element
{
   struct Peelseq_Head next;
   Ind  *sire;
   Ind  *dam;
   Ind  *pivot, *pivot1;
   Indlist *off;
   int   out_index;
   int   n_rfuncs;
   int  *index;
};

struct Complex_Element
{
   struct Peelseq_Head next;
   int   n_peel;                    /* No. alleles to be peeled */
   int   n_involved;                /* No. alleles involved in the operation */
   int   out_index;                 /* Index of output R-Function */
   int   n_rfuncs;                  /* How many R-Functions to be combined */
   int  *involved;                  /* ids of above (i>0 - maternal allele of i,
                                       i<0 - paternal allele of -i) */
   int  *flags;                     /* Flags for involved alleles */
   int  *index;                     /* Indices for R-Functions */
};

/* RFunctions for Mendelian peeling */
struct R_Func
{
   int   n_ind;
   int   n_terms;
   lk_ulong mask[2];
   int   mask1[2];
   int  *id_list;
   double *p;
   int   flag;
   lk_ulong *index;
};

/* RFunctions for polygenic peeling */
struct Poly_R_Func
{
   int   n_ind;
   int   flag;
   int  *type;                      /* Type of included nodes
                                       (additive,dominance,...?) */
   int  *inv;                       /* List of included nodes */
   double *p;
};

struct hash_element
{
   struct hash_element *left, *right;
   lk_ulong index;
   double p;
};

struct hash_block
{
   struct hash_block *next;
   struct hash_element *elements;
   int   size, ptr;
};

GLOBAL int n_markers, n_tloci;
GLOBAL double *ym;                         /* quantitative trait phenotypes */
GLOBAL struct Locus *loci;
GLOBAL int peel_trace;

extern char *MMsg;
extern int max_peel_off, n_singles, *singles;
extern int *idflag, *cplist, *cp_size, **samp_blocks, n_samp_blocks, *nfam;
/* extern int **gene_dat1,**gene_dat2,*gene_idx,n_markers,n_tloci; */
/* extern int *gene_idx, n_markers, n_tloci; */
/* gene_idx not used, n_markers, n_tloci defined above */
extern struct Family **id_fam, *famlist, **flist;

int   amdbar_ (int *n, int *pe, int *iw, int *len, int *iwlen,
               int *pfree, int *nv, int *next, int *last, int *head,
               int *elen, int *degree, int *ncmpa, int *w, int *iovflo);

void   setupSampling (void);
void   setSampling (void);
void   initInherit (int);
void   setFGLpeel (int);
double **initPenProb (int);
void   dbgXXGG (void);
void   dbgGDatGG (int);
void   dbgProb (double **);

void  abt (const char *, const int, const char *,...);
double locus_mjrpeel (double, double, int, int, double *, struct Locus *);
double locus_mjrpeel_cmp (int, double, double, int, int, double *, struct Locus *);
void  locus_peel_init (int *, int);
void  locus_peel_alloc (struct Locus *, int, int *);
void  peel_free (void);
void  set_transmission (struct Locus *, int, int);
void  set_simple_trans (struct Locus *, int, int, double, double,
                        double, double);
void  free_peelseq (struct Peelseq_Head *);
void  locus_peel_free (void);

void  set_qtlpen (double);
void  set_penmat ();                           /* EAT June 2006  */
void free_penmat ();                           /* EAT June 2006 */
void  qtlpen (double *, const Ind *, const struct Locus *, double);
void  recessive (double *, const Ind *, const struct Locus *, double);
void  incomplete (double *, const Ind *, const struct Locus *, double);  /* by AWG */
void  liability (double *, const Ind *, const struct Locus *, double);   /* by AWG */
void  ordinal (double *, const Ind *, const struct Locus *, double);     /* by AWG */
void  set_logistic (void);                                                   /* by LT, May 2007 */
void  logisticpen (double *, const Ind *, const struct Locus *, double);     /* by LT, May 2007 */

double set_quant_trait (int);                   /* EAT Jan 2004 */
/* void  alloc_quantdat(int);              EAT Jan 2004 : removed Aug 2004 */

int   num_bits (const int);
double normalize (double *, int);
void  print_orig_allele_id (FILE *, const int);
void  print_sequence (FILE *, struct Peelseq_Head *, struct R_Func *);
void  print_simple_element (FILE *, const struct Simple_Element *,
                            struct R_Func *);
void  print_complex_element (FILE *, const struct Complex_Element *,
                             struct R_Func *);
void  locus_init_fam (void);
void  locus_free_fam (void);
void  calc_order (const int, struct Family **, int *);

void  set_locpeel (int);
void  set_positions (int);
void  set_lsmpl (int,int, int);
double  do_lsmpl (int, int, double **, int);
void  do_rblik (int, int, int, int *, int *, double *, double *, double *,
                double *, double **, double ***, double *, double *);
void  do_fzplik (int, int, int *, int *, double *, double *, double *,
                double *, double **);
void  do_fzplik2 (int, int, int *, int *, double *, double *, double *,
                double *, double **);
void  scor_fzplik (double **, int, int, int *, int *, double *, 
                double *, double *, double *, double **);
void  scor_fzconf (double **, int, int, int *, int *, double *, 
                double *, double *, double *, double **);
double do_mjrlik (int, int, double, double, double, double, int, double *);
void  do_seq_imp (int, int, int, int, double);

void  prnt_mcmcstrt (void);
void  prnt_mcmcend (void);

/* Tansferred to lsample.h from old sample.h, EAT June 2002 */
void  allocate_loci (int, int *, double **, int);
int   check_seg_mode (void);
void  set_seg_mode (int);

/** @} */

#endif
