/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*     parsedrop.h:  header file for Pars/proc_drop()                         */
/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef PARSEDROP_H             
#define PARSEDROP_H

/**
 * @addtogroup GDrop
 * @{
 */

struct tlist_node {                     /*Tlist_node:  linked or              */
    int no;                             /*linked or unlinked/varying position */
    int lnkd;                           /*trait to be simulated               */
    struct tlist_node * p_next;
                  };
typedef struct tlist_node TLIST;
typedef TLIST * TLIST_PTR;
static TLIST_PTR first_tlist_node = NULL;


/*            built by add_prntmap(), add_only_mark_locus(), add_mark_locus() */

struct prntmap_node {                  /*Prntmap_node:                        */
    int chrom_no;                      /*chromosome no.                       */
    int no_mark_sel;                   /*no. markers simulated/selected       */
    int no_mark_map;                   /*no. markers in map                   */
    int no_traits;                     /*no. traits                           */
    struct prntmap_locus * p_locus;
    struct prntmap_node * p_next;
                    };
typedef struct prntmap_node PRNTMAP;
typedef PRNTMAP * PRNTMAP_PTR;
static PRNTMAP_PTR first_prntmap_node[3] = {NULL, NULL, NULL};

/*            step 1 built by add_only_mark_locus(), add_mark_locus(),        */
/*                            add_trait_locus()                               */
/*            step 2          sel_mark_locus(), sel_trait_locus()             */
/*            step 3          shrink_prntmap(), fix_locus_dist()              */

struct prntmap_locus {                 /*Prntmap_locus:                       */
                                       /*  step 1:                            */
    int locus_typ;                     /*is this locus marker or trait?       */
    int no;                            /*marker or trait no.                  */
    int map_mark;                      /*marker for dist or recomb            */
    double dist;                       /*dist to adjacent marker              */
    double m1_dist;                    /*dist to marker 1                     */
                                       /*  step 2:                            */
    int selectd;                       /*relative selected marker no. or 0    */
    int sel_map_mark;                  /*marker no. for sel_dist              */
    double sel_dist;                   /*dist to selected marker              */
                                       /*  step 3:                            */
    double loc_dist;                   /*dist to prev locus for printing map  */
    struct prntmap_locus * p_prev;
    struct prntmap_locus * p_next;
                     };
typedef struct prntmap_locus PDIST;
typedef PDIST * PDIST_PTR;


static int   loc_spac = 7;             /*no. spaces formatted per locus       */
static int   max_line_loc = 11;        /*up to 11 loci per map segment        */
static char  trait_line[82];           /*with 5 lines formatted for ea        */
static char  trait_tic_marks[82];      /*chrom segment                        */
static char  dist_line[3][82] = {"F  ", "M  ", "   "};
static char  mark_tic_marks[82];
static char  mark_line[82];

                                       /*function prototypes                  */

static SIM_PTR     nxt_sim_chrom(int *);
static DIST_PTR    nxt_mapd_trait(int *,int,SIM_PTR),
                   nxt_mapd_mark(int *,int,SIM_PTR);
static UNLNKD_PTR  nxt_unlnkd_trait(int *);
static TLIST_PTR   nxt_trait(int *);
static int         nxt_mark_no(int *,SIM_PTR);
static PDIST_PTR   add_mark_locus(PRNTMAP_PTR,DIST_PTR),
                   add_only_mark_locus(PRNTMAP_PTR),
                   add_trait_locus(PRNTMAP_PTR,DIST_PTR),
                   nxt_prntmap_locus(int *,PRNTMAP_PTR);
static PRNTMAP_PTR nxt_prntmap(int *,int);
static int         test_sim(void),              
                   test_mark_map(SIM_PTR, int *),      
                   test_trait_map(SIM_PTR, int *),
                   test_trait_list(int,int),
                   test_mark_loci(SIM_PTR,int), test_trait_loci(int),
                   test_loc_mindist(PDIST_PTR, double),
                   test_trait_vars(int),        test_trait_genotyp(int),
                   test_trait_dist(PRNTMAP_PTR,PDIST_PTR),
                   test_seeds(void),            test_oseeds(void);
static PRNTMAP_PTR add_prntmap(SIM_PTR,int);
static void        set_alle_frq_deflts(void),
                   fix_locus_dist(PRNTMAP_PTR,PDIST_PTR),
                   prnt_map(int),  
                   prntmap_tr0(int),            prntmap_tr(int, int),
                   prntmap_mrk0(int),           prntmap_mrk(int, int),
                   prntvmap_tr0(int),           prntvmap_tr(int, int),
                   prntmap_tr0b(void),          prntmap_dist(int, int, double),
                   prntmap_line(int, int),
                   prnt_mark_freqs(SIM_PTR),
                   prnt_trait_freqs(PRNTMAP_PTR),
                   prnt_trait_pars(PRNTMAP_PTR),
                   prnt_unlnkd_freqs(void),
                   prnt_unlnkd_pars(void),
                   fmt_trait_freqs(char *,int,int),
                   fmt_trait_pars(char *,int),
                   alloc_recomb(void),          alloc_trait(void),
                   mak_trait(int,int,int),
                   mak_recomb(PDIST_PTR,int,int),
                   mak_no_recomb(UNLNKD_PTR,int,int),
                   mak_freq(PRNTMAP_PTR,PDIST_PTR,int,int);
static void        for_chrom(int),
                   for_chrom_mark(int,int);

/** @} */

#endif  /*PARSEDROP_H*/
