/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   */
void calc_raoI (int nummrks, int rao_freq, double *rao_realiz)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   */
     /* Version 1.1:  outputing rao_realiz                                    */
     /* This is my first attempt at coming up with a Rao-Blackwellize         */
     /* estimator of P(\lamba_T | Y)/P(\lamba_0 | Y).                         */
     /* The estimator is 

        pi(\lamba_T | Y)        
        --------------------   = 
        pi(\lamba_0 | Y)          

        E_P [  P(Y_T | S_M, lamba_T)P(lamba_T) / sum_{lamba_T)   
        ------------------------------------------------------ X 
        E_P [  P(Y_T | S_M, lamba_0)P(lamba_0) / sum_{lamba_T)     

        P(Y_T | S_M, lamba_T)P(lamba_T)  |  Y]
        ---------------------------------------------
        P(Y_T | S_M, lamba_T)P(lamba_T)  |  Y]
      */
{

   /* Variable declaration                                                    */
   int   sx;
   int   ii, jj;
   int  *locusordr_cpy;                 /* copy of the global chromo where 
                                           chromo[0][k] = locusordr_cpy[k]    */

   double **locusrec_cpy;               /* copy of the global xecomo where
                                           xecomo[sx][0][k] = 
                                           locusrec_cpy[k]    */
   double nc;                           /* normalization constant             */
   double recenter;

   locusrec_cpy = calloc (sizeof (double), 2);
   alloc_chk (WHERE, locusrec_cpy);
   for (sx = FEM; sx <= MAL; sx++)
   {
      locusrec_cpy[sx] = calloc (nummrks + 1, sizeof (double));
      alloc_chk (WHERE, locusrec_cpy[sx]);
   }

   locusordr_cpy = calloc (sizeof (int), nummrks + 1);
   alloc_chk (WHERE, locusordr_cpy);

   /* make a copy of the globals so that they can be restored upon exit       */
   for (sx = FEM; sx <= MAL; sx++)
      for (jj = 0; jj <= nummrks; jj++)
         locusrec_cpy[sx][jj] = xecomo[sx][0][jj];

   for (jj = 0; jj <= nummrks; jj++)
      locusordr_cpy[jj] = chromo[0][jj];

   /* Calculate log P(Y_T | S_M, \lamba_T) across ALL test positions 
      (including unlinked)                                         */
   for (ii = 0; ii < NTPos1; ii++)      /* Remember that NTPos1 is 1 + the 
                                           number of test positions         */
   {
      for (sx = FEM; sx <= MAL; sx++)
         set_chromosomes (nummrks, Marek[sx], 0, &IntVal[ii],
                          &RecVal[sx][ii], sx);
      set_positions (nummrks);          /* assign loci for L sampler          */
      set_transmission (loci, 0, nummrks + 1);
      rao_realiz[ii] = locus_mjrpeel (1.0, 1.0, 0, 0, NULL, loci);
      /* peel only    */
      /*   checks for Zero probability in peeling operation                   */
      if (rao_realiz[ii] == -DBL_MAX)
         give_up_where (WHERE, ERS_zpp);
   }
   /* To minimize over/under flow errors, recenter values                     */

   recenter = rao_realiz[0] + log (PseudoPrior[0]);

   for (ii = 0; ii < NTPos1; ii++)
      rao_realiz[ii] = exp (rao_realiz[ii] + log (PseudoPrior[ii]) - recenter);

   /* Calculate \sum_{\lamba_T} [P(Y_T | S_M, \lamba_T) p(\lamba_T)] for 
      R.B. estimator                                                          */
   nc = 0.0;
   for (ii = 0; ii < NTPos1; ii++)
      nc += rao_realiz[ii];

   for (ii = 0; ii < NTPos1; ii++)
      rao_realiz[ii] = rao_realiz[ii] / nc;

   /* Incremental expectation                                                 */
   for (ii = 0; ii < NTPos1; ii++)
      raoI_est[ii] = (rao_freq - 1) * raoI_est[ii] / rao_freq
         + rao_realiz[ii] / rao_freq;

   /* restore globals and loci                                                */
   for (sx = FEM; sx <= MAL; sx++)
      for (jj = 0; jj <= nummrks; jj++)
         xecomo[sx][0][jj] = locusrec_cpy[sx][jj];

   for (jj = 0; jj <= nummrks; jj++)
      chromo[0][jj] = locusordr_cpy[jj];

   set_positions (nummrks);             /* restore loci */

   /* free memory                                                             */

   for (sx = FEM; sx <= MAL; sx++)
      free ((void **) locusrec_cpy[sx]);
   free ((void *) locusrec_cpy);
   free ((void *) locusordr_cpy);

   return;
}
