/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

static int *perm_sw = NULL;             /* permutation arrays of switches     */
static int *perm_loc = NULL;            /* and loci                           */

/*  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                 */
void update_loci (double prob_L,
                  int nummrks, int nped, int nsw, double **prob, int ng)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                 */
/* Uses the L sampler or M sampler to update all loci and meioses             */
/*    respectively via Gibbs proposals.                                       */
/* nsw - number of meioses/switches                                           */
/* ng  - number of genes at any given locus - based on pedigree structure     */
{
   int   i, j, l, sx;
   int   c;                             /* component index  */
   int   cloc;                          /* locus being update  */
   double rnum;                         /* random number (0-1)  */
   double prb_Yt_cndtl;                 /* log Pr(Y_T | S_T-1, S_T+1)  */

   /* set up permutation arrays for loci and switches                         */
   if (!perm_sw)
   {
      perm_sw = malloc (nsw * sizeof (int));
      alloc_chk (WHERE, perm_sw);
      perm_loc = malloc ((nummrks + 1) * sizeof (int));
      alloc_chk (WHERE, perm_loc);
   }
   for (i = 0; i < nsw; ++i)
      perm_sw[i] = i;
   for (i = 0; i <= nummrks; ++i)
      perm_loc[i] = i;

   rnum = rans ();
   if (rnum <= prob_L)
   {

      /* +-------------------------+                                          */
      /* |     L SAMPLER           |                                          */
      /* +-------------------------+                                          */

      permute (perm_loc, nummrks + 1);
      for (j = 0; j <= nummrks; j++)
      {
         cloc = perm_loc[j];
         if (cloc != 0)                 /* Update the trait locus last */

            prb_Yt_cndtl = do_lsmpl (nummrks, cloc, prob, ng);
      }

      /* Update the trait locus and calculate
         log P(Y_T | S_i, S_i+1, \lamba, \alpha)                              */

      prb_Yt_cndtl = do_lsmpl (nummrks, 0, prob, ng);

      previous_realiz.prb_Yt_condtl = prb_Yt_cndtl; /* log P(Y_T | S_M) */

   }

   else
   {

      /* +-------------------------+                                          */
      /* |     M SAMPLER           |                                          */
      /* +-------------------------+                                          */

      permute (perm_sw, nsw);
      for (j = 0; j < nsw; j++)
      {
         l = perm_sw[j] + 1;
         i = (meiosis[l].ind).index;
         sx = meiosis[l].sex;
         c = nghd[i].component - 1;

         do_msmpl (nummrks, 0, c, i, sx, prob, ng);
      }

      /* Calculate log Pr(Y_T | S_{T-1}, S_{T+1}) and why not update S_T 
         at the same time                                                     */

      prb_Yt_cndtl = do_lsmpl (nummrks, 0, prob, ng);

      previous_realiz.prb_Yt_condtl = prb_Yt_cndtl; /* log P(Y_T | S_M) */
   }

   /* Store   xx0 and  xx1 in previous_realiz                                 */

   for (i = 0; i <= nummrks; i++)
      for (j = 0; j < nped; j++)
      {
         previous_realiz.xx0[i][j] = xx0[i][j];
         previous_realiz.xx1[i][j] = xx1[i][j];
      }
   return;

}
