/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*******************************************************************************
  program name: lm_multiple.c (Estimate or calculate LodScore by component. 
                Components can be input in any order)

 (LPT Jul 2007)

  Note that this program is modified and very similar to lm_markers.c.
  The differences are that 
  (1) this program allows exact calculation of LodScores for a small pedigree 
      with no. of meiosis less than or equal to MAXNMexact, which is a constant 
      defined in header file /Headers/mulmsample.h.
  (2) this program performs additional multiple meioses updating: complete and 
      restricted individual, sib and 3-generation updating. The constant 
      MAXNMupdate, defined in /Headers/mulmsample.h, determines wether to 
      perform a complete updating or a restricted one. The constants PMEIO,
      PIND, PSIB, P3GEN are probabilities for each type of updating. They
      are defined in /Headers/mulmsample.h and their sum must be 1.0.

  The state (X) consists of segregation indicators -- so can generalize to
        cases of more than one observed/pedigree, and multiple loci.

  The data, Y, is the multilocus genotypes of observed individuals
     in the pedigree.

  Obtained by modification of jnt_auto.c.  December 1998.  (EAT)

Here is the plan!
    OK this is the PLAN  (12/15/98)

OK, now we have a general autozyg program, we need to go back to
lod estimation, for those as want.  We use the same program (modified).

Since the data are locus-by-locus, the likelihood ratio contribution is
just ratio of P_theta(S)/P_sim_th(S).

This is very easy, once we get the chromosome set-ups fixed.

1) Trait locus data set-up -- assume trait-locus genotypes for observed
individuals -- done  12/20/98
2) Likelihood-ratio contributions routine,  -- done 12/18/98
3) Simulation and evaluation chromosomes for each location. -- done 12/23/98

Now to integrate the L-sampler, using the new Lsampler and Msampler
      libraries and new routines;        EAT March 2000

Converted to MORGAN_V2.3                 MAJ June 2000

Modifications to specify simulation points, within each interval and
outside of map.    EAT, then converted to MORGAN 2.3 by MAJ, Nov 2000.

Modification of full_seg_prob, and recombinant counts     EAT  Nov 2000
    scoring of counts and tot_log_probs by component and locus/interval.

Modifications to lik_rat, do_lsmpl, etc.                  EAT  Dec 2000

Modifications to use specify intervals for l-ratio updating and use
    more efficient seg_prob updating routines for l-sample steps
                                                          EAT Jan 2001

Added changes for sequential imputation set-up            AWG March 2001

But now we can use locus_mjrpeel with sample_flag zero
     to do a Lange-style lod score using only sampled S_M    EAT May 2001
  We can use the extval and intval positions as computation positions
    -- except we can also compute at the loci, maybe??

This is the version to test new trait stuctures.
Program renamed lm_markers.                                 EAT March 2002

Version modified for quantitative traits (lm_quant.c)       EAT Jan 2004

Output scores file added for CDLL output                    EAT Apr 2004

Exact calculation for small pedigrees added                 LPT Feb 2006

Multiple meioses updating for MCMC added                    LPT Apr 2006

Multiple meiosis updating and exact calculation cleaned up  EAT Sep 2006
  Dependence on new MulMSample versions of setupAutozyg()
   and setupMaps() is removed: these shd now be obsolete.

*******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_markers"
#define MARKERS

#include "lodscore.h"

int main (int argc, char *argv[])
{
   int   nCompnt1;
   double *temp_lod, *temp_sd, *overall_lod, *overall_sd;
   int   c, k;
   int   ntp, mul;                      /* mul is indicator of MM-sampling   */
   int   nsw;

   double **prob;

   /* Map position (cM) of current test position */
   double **rec_left;                   /* fem =0, male=1; by test position   */
   double **rec_rght;                   /* fem =0, male=1; by test position   */
   int  *loc_left;                      /* locus left of test position        */
   int  *loc_rght;                      /* locus rght of test position        */
   FILE *oSeedF;

/* set up for multiple-meiosis sampling or old M-sampler   */

   mul = 1;
   Max_update = 8;
   P_meio = 0.2;
   P_ind = 0.3;
   P_sib = 0.3;
   P_3gen = 0.2;
#ifdef NO_MUL
   mul = 0;
   Max_update = 1;
   P_meio = 1.0;
   P_ind = 0.0;
   P_sib = 0.0;
   P_3gen = 0.0;
#endif

   /* .................................................. */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();

   nCompnt1 = 0;
   if (!GlobalMcmc)
   {
      /* number of components to be calculated exactly */
      for (c = 1; c <= nCompnt; c++)
         if (cmplist[c].nmeioses <= MAXNMexact)
            nCompnt1++;
   }

   /* setupInherit moved down, so it knows whether exact comp will be done */

   if (!nCompnt1)
      prob = setupInherit (1);
   else
      prob = setupInherit (0);

   /* now set evaluation points for lange_lik */
   /* ntp is number of trait positions plus number of markers (NLocPos + 1) */

   alloc_lf_rt (NLocPos + 1, &rec_left, &rec_rght, &loc_left, &loc_rght);

   ntp = set_eval_points (NMarks, loc_left, loc_rght,
                          rec_left[0], rec_rght[0], rec_left[1], rec_rght[1]);

   if (ntp != NLocPos + 1)
      give_up_where (WHERE, ERA_mrk, NLocPos, ntp);

   /* .................................................. */

   /* now estimate lods using mcmc for large pedigrees 
      and find exact lods for small ones */

   al_double_vector (&temp_lod, ntp);
   al_double_vector (&temp_sd, ntp);
   al_double_vector (&overall_lod, ntp);
   al_double_vector (&overall_sd, ntp);
   for (k = 1; k < ntp; k++)
      overall_lod[k] = overall_sd[k] = 0;

   if (mul)
   {
      setup_marriage ();
      prnt_marriage ();
   }

   /* For global MCMC, use mcmc_lods with first parameter =0 */
   if (GlobalMcmc)
   {
      mcmc_lods (0, NBrnIt, NIt, mul, ntp, rec_left, rec_rght, loc_left,
                 loc_rght, prob, temp_lod, temp_sd);

      printf ("\n\n === LodScore for all the components completed ===");
      printf ("\n\n");
      fflush (stdout);

      prntLodSHead ();
      for (k = 1; k < ntp; ++k)
      {
         if ((!prntLodSPos (k - 1)) && (TraitTyp == GENOTYPIC))
            /* the current test position is a marker position.
               This test is reinstated: EAT Oct 2005.   It is needed for
               GENOTYPIC traits in some cases.  See note re do_rblik() */
            printf (" %11s%10s\n", "NA", "NA");
         else
            printf (" %11.4f%10.4f\n", temp_lod[k], temp_sd[k]);
      }
      printf ("\n");
      fflush (stdout);
   }
   else
   {
      for (c = 1; c <= n_compnts; c++)
      {
         if (cmplist[c].nmeioses <= MAXNMexact)
         {
            printf
               ("\n\n ========= LodScore for Component %d is calculated exactly =========\n\n",
                c);

            exact_lods (c, NMarks, nFGL, ntp, temp_lod,
                        loc_left, rec_left[1], rec_left[0],
                        loc_rght, rec_rght[1], rec_rght[0]);
            for (k = 1; k < ntp; k++)
               overall_lod[k] += temp_lod[k];

            printf ("\n Summary for component %d:", cmplist[c].index);
            printf ("\n Base log-likelihood log_10 P(Y_T) = %e\n \n",
                    temp_lod[0]);
            fflush (stdout);

            printf (" LodScore by exact computation:\n\n");
            printf (" Trait pos #     position (Haldane cM)\n");
            printf ("   or marker        male     female    %11s%10s",
                    "LodScore", "StdErr");
            printf ("\n\n");
            fflush (stdout);

            for (k = 1; k < ntp; k++)
            {
               if ((!prntLodSPos (k - 1)) && (TraitTyp == GENOTYPIC))
                  /* the current test position is a marker position.
                     This test is reinstated: EAT Oct 2005.   It is needed for
                     GENOTYPIC traits in some cases.  See note re do_rblik()
                     Test is added here, EAT, Aug 2008                     */
                  printf (" %11s%10s\n", "NA", "NA");
               else
                  printf (" %11.4f%10.4f\n", temp_lod[k], 0.);
            }
            printf ("\n");
            fflush (stdout);
         }
         else
         {
            printf
               ("\n\n ========= LodScore for Component %d is estimated using MCMC =========\n\n",
                c);
            mcmc_lods (c, NBrnIt, NIt, mul, ntp, rec_left, rec_rght, loc_left,
                       loc_rght, prob, temp_lod, temp_sd);

            for (k = 1; k < ntp; ++k)
            {
               overall_lod[k] += temp_lod[k];
               overall_sd[k] += temp_sd[k];
            }

            prntLodSHead ();
            for (k = 1; k < ntp; ++k)
            {
               if ((!prntLodSPos (k - 1)) && (TraitTyp == GENOTYPIC))
                  /* the current test position is a marker position.
                     This test is reinstated: EAT Oct 2005.   It is needed for
                     GENOTYPIC traits in some cases.  See note re do_rblik() */
                  printf (" %11s%10s\n", "NA", "NA");
               else
                  printf (" %11.4f%10.4f\n", temp_lod[k], temp_sd[k]);
            }
            printf ("\n");
            fflush (stdout);
         }
      }

      /* now print the overall cumulative lodscore */

      if (nCompnt > 1)
      {
         printf ("\n\n === LodScore for all the components completed ===");
         printf ("\n\n");
         fflush (stdout);

         printf ("\n LodScore estimates through all components:\n\n");
         printf (" Trait pos #     position (Haldane cM)\n");
         printf ("   or marker        male     female    %11s%10s", "LodScore",
                 "StdErr");
         printf ("\n\n");
         fflush (stdout);

         for (k = 1; k < ntp; ++k)
         {
            if ((!prntLodSPos (k - 1)) && (TraitTyp == GENOTYPIC))
               /* the current test position is a marker position.
                  This test is reinstated: EAT Oct 2005.   It is needed for
                  GENOTYPIC traits in some cases.  See note re do_rblik() */
               printf (" %11s%10s\n", "NA", "NA");
            else
               printf (" %11.4f%10.4f\n", overall_lod[k], overall_sd[k]);
         }
         printf ("\n");
         fflush (stdout);
      }
   }

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Free up allocated memory */
   free_marriage ();

   free_1d ((void *) &temp_lod);
   free_1d ((void *) &temp_sd);
   free_1d ((void *) &overall_lod);
   free_1d ((void *) &overall_sd);

   free_xxgg ();
   free_allefrq ();
   free_chromo (0);
   free_allocate_loci ();
   free_perm ();
   locus_peel_free ();

   /* now we are done  */

   exit (EXIT_SUCCESS);
}
