/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/* set_markdat(), set_traitdat(), set_probgam()               MAJ Aug/01   */
/* moved to Msample library.                                  EAT Feb 02   */
/* set_markdat() moved to set_traits.c                        EAT Mar 02   */
/* set_affind() added                                         MAJ MAY 03   */
/* printout from set_traitdat() moved to prnt_traitdat()      MAJ JAN 04   */
/* edited to change traits storage etc.                       EAT AUG 04   */
/* edited various to get rid of genotype array y              EAT SEP 04   */
/* updating of set_affind()                                   EAT Dec 04   */
/* Note this file is now primarily trait set-up -- will move to trait lib  */

#include "markers.h"

void set_traitdat (int pdsz, int nc, int nd)
{
   int   i, np1;
   int  *pdat;

/* if trait is GENOTYPIC, the data is now set with the marker data
   in set_markdat   EAT Sep 2004
 */

   if (PedDatTyp != DISCRETE && PedDatTyp != DISCR_LIAB
       && PedDatTyp != DISCR_AGE)
      return;

   np1 = 0;
   for (i = 0; i < pdsz; ++i)
      if (data[i][5] != 0)
         ++np1;

   pdat = malloc (nd * np1 * sizeof (int));
   alloc_chk (WHERE, pdat);

   for (i = 0; i < pdsz; ++i)
   {
      if (data[i][5] != 0)
      {
         nghd[i].d_dat = pdat;
         nghd[i].d_dat[0] = data[i][5];
         if (nd == 2)
            nghd[i].d_dat[1] = data[i][6];
         nghd[i].obs = 1;
         pdat += nd;
      }
   }

   return;
}

void prnt_traitdat (int pdsz, int nc)
{
   int   c, k, i, ifound, kfound;

   char *k_txt[5] = {
      "",
      "(homozygous normal)",
      "(not implemented)",
      "(heterozygote)",
      "(homozygous affected)"
   };

   if (TraitTyp != GENOTYPIC && TraitTyp != DISCRETE)
      return;

   printf (" Trait data:\n");

   for (c = 0; c < nc; ++c)
   {
      kfound = 0;
      printf ("\n Component %d:\n", c + 1);
      for (k = 4; k > 0; --k)           /* k=0 unknown        */
      {
         ifound = 0;

         for (i = 0; i < pdsz; ++i)
         {
            if (nghd[i].component == c + 1 && data[i][5] == k)
            {
               if (TraitTyp == DISCRETE)
               {
                  kfound = 1;
                  if (!ifound)
                     printf ("\n    phenotype %d:", data[i][5]);
                  if (!(ifound % 6))
                     printf ("\n   ");
                  ++ifound;
                  printf (" %s", nghd[i].self);
               }
               else if (TraitTyp == GENOTYPIC)
               {
                  if (k == 4 || k == 3 || k == 1)
                     kfound = 1;
                  if (!ifound)
                     printf ("\n    genotype %d %d %s:",
                             nghd[i].g_dat[0], nghd[i].g_dat[1], k_txt[k]);
                  if (!(ifound % 6))
                     printf ("\n   ");
                  ++ifound;
                  printf (" %s", nghd[i].self);
               }
            }
         }
         if (ifound)
            printf ("\n");
      }
      if (!kfound)
         printf ("    no trait data\n");
   }
   printf ("\n\n");
   fflush (stdout);

   return;
}

int set_probgam (int nc)
{
   int   c, s, k, i;
   int   npc = 0;

   for (c = 0; c < nc; ++c)             /* for each component  */
      if (NScCmp < c + 1)
         give_warn (ERA_ncd, c + 1);

      else if (!NScSet[c])
         give_warn (ERA_ncd, c + 1);

      else
      {
         ++npc;
         for (s = 0; s < NScSet[c]; ++s) /* and each scoreset                */
            for (k = 0; k < NPrGam[c][s]; ++k)
            {                           /* gets indices to data[][]           */
               if ((i = get_posit (ProbNam[c][s][k])) == INVALID_INDEX)
                  give_up (ERA_inp, ProbNam[c][s][k]);
               else if ((nghd[i].component) != (c + 1))
                  give_up (ERA_pnc, ProbNam[c][s][k], c + 1, nghd[i].component);

               ProbIx[c][s][k] = i;
            }
      }
   if (!npc)                            /* checks for any proband gametes     */
      give_up_where (WHERE, ERA_mpg);
   else if (NScCmp > nc)
      give_warn (ERA_mcd, NScCmp, nc);

   return nc < NScCmp ? nc : NScCmp;
}

void set_affind (int nc, int a_only)
/* iff a_only = TRUE, sets only affecteds; else both affected & unaffected */
{
   int   c, i, k;
   int   nac = 0;
#if 0
   for (i = 0; i < pedsize; ++i)
      /* added EAT 01/2005: apparently needed for both AffTrIn and !AffTrIn */
      /* locus 0 data must be zeroed  :  TraitTyp will be (re)set GENOTYPIC */
      if (nghd[i].g_dat)
      {
         nghd[i].g_dat[0] = 0;
         nghd[i].g_dat[1] = 0;
      }
#endif

   if (AffTrIn)                         /* Affecteds to be found in ped file  */
   {
      if (!((TraitTyp == GENOTYPIC) || (TraitTyp == DISCRETE)))
      {
         printf (" PROBLEM: your trait has type %d\n", TraitTyp);
         printf (" NTraits =  %d\n", NTraits);
         give_up_where (WHERE, ERR_err, "Trait must be genotypic or discrete");
      }

      NAffCmp = nc;

      for (c = 0; c < nc; ++c)
      {
         cmplist[c + 1].naff = 0;
         if (!a_only)
            cmplist[c + 1].nunaff = 0;
      }

      for (i = 0; i < pedsize; ++i)     /* counts no. affecteds in ea compnt  */
      {
         if (TraitTyp == GENOTYPIC && nghd[i].g_dat)
         {
            if (nghd[i].g_dat[1] == 2)  /* (1,2) or (2,2) */
            {
               ++cmplist[nghd[i].component].naff;
               ++nac;
            }
            if (nghd[i].g_dat[1] == 1)  /* (1,1) */
               if (!a_only)
                  ++cmplist[nghd[i].component].nunaff;
         }
         else if (TraitTyp == DISCRETE && nghd[i].d_dat)
         {
            if (nghd[i].d_dat[0] == 2)  /* affected */
            {
               ++cmplist[nghd[i].component].naff;
               ++nac;
            }
            else if (nghd[i].d_dat[0] == 1) /* unaffected */
               if (!a_only)
                  ++cmplist[nghd[i].component].nunaff;
         }
      }

      for (c = 0; c < nc; ++c)
         printf (" Now set %d affected and %d unaffected in component %d\n",
                 cmplist[c + 1].naff, cmplist[c + 1].nunaff, c + 1);

      for (c = 0; c < nc; ++c)          /* for each component: */
      {
         if (cmplist[c + 1].naff)
         {
            cmplist[c + 1].ind_aff =
               malloc (cmplist[c + 1].naff * sizeof (int));
            alloc_chk (WHERE, cmplist[c + 1].ind_aff);

            k = 0;
            for (i = 0; i < pedsize; ++i)
            {
               if (data[i][5] != 0 && nghd[i].component == c + 1)
               {
                  if ((TraitTyp == GENOTYPIC) &&
                      (data[i][5] == 3 || data[i][5] == 4))
                  {
                     cmplist[c + 1].ind_aff[k] = i; /* sets index   */
                     ++k;
                  }
                  if ((TraitTyp == DISCRETE) && (data[i][5] == 2))
                  {
                     cmplist[c + 1].ind_aff[k] = i; /* sets index   */
                     ++k;
                  }
               }
            }
            if ((!a_only) && (cmplist[c + 1].nunaff))
            {
               cmplist[c + 1].ind_unaff =
                  malloc (cmplist[c + 1].nunaff * sizeof (int));
               alloc_chk (WHERE, cmplist[c + 1].ind_unaff);

               k = 0;
               for (i = 0; i < pedsize; ++i)
               {
                  if ((data[i][5] == 1) && (nghd[i].component == c + 1))
                  {
                     cmplist[c + 1].ind_unaff[k] = i; /* sets index   */
                     ++k;
                  }
               }
            }
         }
         else
         {
            give_warn (ERA_nff, c + 1);
         }
         if (!nac)
            give_up (ERA_nap, nc);
         /*  print_affind (); */
      }
      TraitTyp = UNDEF;                 /* TraitTyp for AffTrIn is also UNDEF */
   }
   else if (a_only)
   {                                    /* Affecteds in parameter file        */
      for (c = 0; c < nc; ++c)          /* for each component:                */
      {
         if (NAffCmp < c + 1)
            give_warn (ERA_naf, c + 1);

         else if (!NAffInd[c])
            give_warn (ERA_naf, c + 1);
         else
         {
            ++nac;
            cmplist[c + 1].naff = NAffInd[c];
            cmplist[c + 1].ind_aff =
               malloc (cmplist[c + 1].naff * sizeof (int));
            alloc_chk (WHERE, cmplist[c + 1].ind_aff);
            for (k = 0; k < cmplist[c + 1].naff; ++k)
               /* gets the data[][] indices          */
            {
               if ((i = get_posit (AffNam[c][k])) == INVALID_INDEX)
                  give_up (ERA_nip, AffNam[c][k]);
               else if ((nghd[i].component) != (c + 1))
                  give_up (ERA_nic, AffNam[c][k], c + 1, nghd[i].component);

               cmplist[c + 1].ind_aff[k] = i;
            }
         }
      }
      if (NAffCmp > nc)
         give_warn (ERA_mca, NAffCmp, nc);
      if (!nac)
         give_up (ERA_naa, nc);
      print_affind ();

      if (AffNam)
      {
         for (c = 0; c < NAffCmp; ++c)
         {
            for (i = 0; i < NAffInd[c]; ++i)
               free ((void *) AffNam[c][i]);
            AffNam[c] = NULL;
            free ((void *) AffNam[c]);
         }
         free ((void *) AffNam);
         AffNam = NULL;
      }
   }
   else
   {
      printf (" PROBLEM: Only affecteds can be specified"
              " in the parameter file\n");
      give_up_where (WHERE, ERR_err,
                     " If you require both Aff and Un-aff,"
                     " specify as phenotype\n     in the pedigree file");
   }

   for (i = 0; i < pedsize; ++i)
      /* added EAT 01/2005: apparently needed for both AffTrIn and !AffTrIn */
      /* locus 0 data must be zeroed  :  TraitTyp will be (re)set GENOTYPIC */
      if (nghd[i].g_dat)
      {
         nghd[i].g_dat[0] = 0;
         nghd[i].g_dat[1] = 0;
      }

   return;
}

/* print_affind moved here                              EAT 12/2004   */
/* since it is now called with the cmplist structures                 */

 /* prints, for each component, the list of affected individuals.
  */
void print_affind (void)
{
   int   c, i, n, nn;
   int   nc = 0;

   for (c = 0; c < NAffCmp; ++c)
   {
      for (i = 0; i < cmplist[c + 1].naff; ++i)
      {
         n = strlen (nghd[cmplist[c + 1].ind_aff[i]].self);
         /* gets no. chars in names for        */
         nc = n > nc ? n : nc;          /* printing                           */
      }
   }
   nn = 76 / (nc + 1);

   for (c = 0; c < NAffCmp; ++c)
   {
      for (i = 0; i < cmplist[c + 1].naff; ++i)
         /* prints list of affected individs   */
      {
         if (!i)
         {
            fprintf (stdout, " Component %d:\n\n", c + 1);
            fprintf (stdout, " %d affected individuals:", cmplist[c + 1].naff);
         }
         if (!(i % nn))
            fprintf (stdout, "\n   ");
         fprintf (stdout, " %*s", nc, nghd[cmplist[c + 1].ind_aff[i]].self);
         if (i == cmplist[c + 1].naff - 1)
            fprintf (stdout, "\n\n");
      }
   }
   fflush (stdout);

   return;
}
