/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/* These functions are called by the parameter statement functions, appearing
 * in proc_pars.c and named in parsedefs.h
 */

/* ---- check fncs:  check contents of parameter statements for errors ----- */

int chk_stat (void)
{                                       /* increments counter for this       */
   ++ppp->count;                        /* statement type                    */

   if (strchr (ppp->pcods, pmx->pcod))  /* checks for this statement         */
      return TRUE;                      /* implemented for program           */
   else
   {
      err_typ = ERR_NOT;
      err_txt = pmx->pnam;
      return FALSE;
   }
}
int chk_repeat (void)                   /* test for redundant statement      */
{
   if (ppp->count > 1)
   {
      err_typ = ERR_RED;
      return FALSE;
   }
   else
      return TRUE;
}
void chk_fil (char *f_typ)              /* tests for file name already given */
{                                       /* stores file name                  */
   pft = get_filent (f_typ);

   if (!pft)
      return;

   if (pft->command)
   {
      err_typ = ERR_CFO;
      sprintf (errnumb, "%s", pft->fdesc);
      err_txt = errnumb;
   }
   else
   {
      if (!chk_repeat ())
         free_fnam (pft);
      pft->fname = alloc_copy (s_par);
      pft->done = N;
      pft->rewrit = N;
   }
   pft->param = Y;

   return;
}
void set_rewrt (char *f_typ)
{
   pft = get_filent (f_typ);            /* sets overwrite file option */

   if (!pft)
      return;

   pft->rewrit = Y;

   return;
}
int chk_trait (void)                    /* tests for qualitative trait no.   */
                                        /* out of range                      */
{
   if (j_par[0] < 1 || j_par[0] > FMAX_REAL)
   {
      err_typ = ERR_TRO;
      return FALSE;
   }
   else
      return j_par[0];
}
int chk_int_trait (void)                /* tests for trait no. out of range  */
                                        /* out of range                      */
{
   if (j_par[0] < 1 || j_par[0] > FMAX_INT)
   {
      err_typ = ERR_TRO;
      return FALSE;
   }
   else
      return j_par[0];
}

int chk_trtped (int typ)                /* tests list of trait no.'s vs pos   */
{                                       /* for integers, int pairs, or reals  */
   int   m, i;

   m = typ == INTDAT2 ? 3 : 2;

   if (m == 2)
   {
      if (nk_par % 2)
         err_typ = ERR_WTI;
   }
   else if (nk_par % 3)
      err_typ = ERR_WT2;
   if (err_typ)
      return FALSE;

   if (typ != REALDAT)                  /* checks record postitions           */
   {
      for (i = nk_par / m; i < nk_par; ++i)
         if (j_par[i] < 1 || j_par[i] > FMAX_INT)
         {
            err_typ = ERR_TRI;
            return TRUE;
         }
   }
   else
      for (i = nk_par / m; i < nk_par; ++i)
         if (j_par[i] < 1 || j_par[i] > FMAX_REAL)
         {
            err_typ = ERR_TRK;
            return TRUE;
         }

   for (i = 0; i < nk_par / m; ++i)     /* checks trait numbers               */
      if (j_par[i] < 1 || j_par[i] > FMAX_INT)
      {
         err_typ = ERR_TRO;
         return TRUE;
      }
   return TRUE;
}

TRTPED *chk_trtped_node (int no)        /* tests for record position already  */
                                       /* already given for this trait        */
{
   TRTPED_PTR p_node;

   p_node = first_trtped_node;
   if (p_node)
   {
      while (p_node->p_next && p_node->no < no)
         p_node = p_node->p_next;
      if (p_node->no == no)
      {
         err_typ = ERR_TRP;
         return p_node;
      }
   }
   return NULL;
}

int chk_sel_tr (void)                   /* tests for repeated trait no.      */
{
   if (sel_qtrait[j_par[0] - 1])
   {
      err_typ = ERR_TRR;
      return FALSE;
   }
   else
      return j_par[0];
}
int chk_ff (int m)                      /* tests for fixed effect no.        */
{                                       /* out of range                      */
   int   k;

   if (!m)                              /* select ...                        */
   {
      for (k = 1; k <= nk_par; ++k)
         if (j_par[k] < 1 || j_par[k] > FMAX_FEF)
         {
            err_typ = ERR_FFO;
            return FALSE;
         }
      return TRUE;
   }
   else if (j_par[1] < 1 || j_par[1] > FMAX_FEF) /* start ...               */
   {
      err_typ = ERR_FFO;
      return FALSE;
   }
   else
      return TRUE;
}
int chk_start_ff (int m)                /* tests for mean or fixed effect    */
{                                       /* levels already given for trait    */

   if (!m)                              /* mean                              */
   {
      if (nstartff[j_par[0] - 1][0] == 1)
      {
         err_typ = ERR_MNR;
         return FALSE;
      }
      else
         return TRUE;
   }
   else
   {
      if (nstartff[j_par[0] - 1][j_par[1]] != 0) /* fixed effect            */
      {
         err_typ = ERR_FFR;
         return FALSE;
      }
      else
         return TRUE;
   }
}
int chk_levels_ff (void)                /* tests for too many starting       */
{                                       /* levels for fixed effect           */
   if (nx_par > LEVELS)
   {
      err_typ = ERR_LVM;
      return FALSE;
   }
   else
      return TRUE;
}
int chk_var (void)                      /* tests for too many starting       */
{                                       /* values for variances              */
   if (nx_par > pmx->max_trait)
   {
      err_typ = ERR_VAM;
      return FALSE;
   }
   else
      return TRUE;
}
int chk_covar (void)                    /* tests for too many starting       */
{                                       /* values for covariances            */
   if (nx_par > pmx->max_trait * (pmx->max_trait - 1) / 2)
   {
      err_typ = ERR_COM;
      return FALSE;
   }
   else
      return TRUE;
}
int chk_covar_i (void)                  /* tests for too many fit flags for  */
{                                       /* covariances                       */
   if (nk_par > pmx->max_trait * (pmx->max_trait - 1) / 2)
   {
      err_typ = ERR_CFM;
      return FALSE;
   }
   else
      return TRUE;
}
int chk_sim (int c, int m, int t)       /* tests for chromosome no.,         */
{                                       /* no. markers, no. traits, out       */
   int   i;                             /* of range                           */
   int   ix = 0;

   if (CH == c)
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_chrom)
      {
         err_typ = ERD_CHO;
         return FALSE;
      }
      ix++;
   }
   if (MK == m)
   {
      if (j_par[ix] < 1)
      {
         err_typ = ERD_MK0;
         return FALSE;
      }
      if (j_par[ix] > pmx->max_mark)
      {
         err_typ = ERD_MKM;
         return FALSE;
      }
      ix++;
   }
   else if (MKL == m)
   {
      for (i = ix; i < nk_par; ++i, ++ix) /* checks marker no's               */
         if (j_par[ix] < 1 || j_par[ix] > pmx->max_mark)
         {
            err_typ = ERD_MKO;
            return FALSE;
         }
   }
   if (TR == t)
   {
      if (j_par[ix] < 0)
      {
         err_typ = ERD_TR0;
         return FALSE;
      }
      if (j_par[ix] > pmx->max_trait)
      {
         err_typ = ERD_TRM;
         return FALSE;
      }
   }
   if (TRL == t)
      ;

   return TRUE;
}

SIM  *chk_sim_node (int c)              /* tests for statement already given */
{                                       /* for this chromosome no.           */
   SIM_PTR p_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];
   if ((p_node = first_sim_node))
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERD_RSM;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

int chk_unlnkd (void)                   /* tests for no. traits and trait    */
{                                       /* no's out of range                 */
   int   i;

   if (nk_par > pmx->max_trait)
   {
      err_typ = ERD_UTM;
      return FALSE;
   }
   for (i = 0; i < nk_par; i++)
      if (j_par[i] < 1 || j_par[i] > pmx->max_trait)
      {
         err_typ = ERD_TRO;
         return FALSE;
      }
   return TRUE;
}

int chk_map (int c, int d, int m)
/*                                       * tests for chromosome no., no. of
 *                                         markers, no. of traits or no. of
 *                                         distances out of range
 *                                         also tests for distances or recomb
 *                                         fractions out of range             */
{
   int   i, ix = 0;
   int   err_ret = FALSE;

   if (CH == c)                         /* checks chromosome no.              */
   {
      if (j_par[0] < 1 || j_par[0] > pmx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ix++;
   }
   if (err_ret)
      return FALSE;

   if (m == MARK_LOC)                   /* checks map marker statement        */
      switch (d)
      {
          case RM:
             if (nx_par >= pmx->max_mark)
             {
                err_typ = ERD_MRM;
                err_ret = TRUE;
             }
             break;
          case DM:
          case KDM:
             if (nx_par >= pmx->max_mark)
             {
                err_typ = ERD_MDM;
                err_ret = TRUE;
             }
             break;
          case PM:
          case KPM:
             if (nx_par > pmx->max_mark)
             {
                err_typ = ERD_MPM;
                err_ret = TRUE;
             }
             else if (nx_par == 1)
             {
                err_typ = ERD_O1M;
                err_ret = TRUE;
             }
             break;
          default:
             break;
      }

   if (err_ret)
      return FALSE;

   if (m == TRAIT_LOC)                  /* checks map trait statement         */
   {
      if (nk_par != 2 * nx_par)
      {
         err_typ = ERD_TMM;
         err_ret = TRUE;
      }
      else if (nx_par > pmx->max_trait)
      {
         if (RM == d)
            err_typ = ERD_MRM;
         else if (DM == d || KDM == d)
            err_typ = ERD_MDM;
         err_ret = TRUE;
      }
      if (!err_ret)
         for (i = 0; i < nk_par / 2; i++)
            if (j_par[ix + i] < 1 || j_par[ix + i] > pmx->max_trait)
            {
               err_typ = ERD_TRO;
               err_ret = TRUE;
            }
      if (!err_ret)
         for (i = 0; i < nk_par / 2; i++)
            if (j_par[ix + nk_par / 2 + i] < 0 ||
                j_par[ix + nk_par / 2 + i] > pmx->max_mark)
            {
               err_typ = ERD_MKO;
               err_ret = TRUE;
            }
   }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nx_par; i++)         /* checks fractions, distances        */
   {                                    /* or positions                       */
      switch (d)
      {
          case RM:
             if (d_par[i] < 0. || d_par[i] >= .5)
             {
                if (m == MARK_LOC)      /* tests for recomb fraction >= 0.    */
                   err_typ = ERD_MFO;   /* and < .5                           */
                else
                   err_typ = ERD_TFO;
                err_ret = TRUE;
             }
             break;
          case DM:
          case KDM:
             if (d_par[i] < 0. || d_par[i] > 500.)
             {
                if (m == MARK_LOC)      /* tests for distance >= 0.           */
                   err_typ = ERD_MDO;   /* and <= 500. cM                     */
                else
                   err_typ = ERD_TDO;
                err_ret = TRUE;
             }
             break;
          case PM:
          case KPM:
             if (d_par[i] < -500. || d_par[i] > 500.)
             {
                err_typ = ERD_MPO;      /* tests for position .= -500.        */
                err_ret = TRUE;         /* and <= 500 cM                      */
             }
             else if (i && d_par[i] < d_par[i - 1])
             {
                err_typ = ERD_MPD;      /* also tests for increasing order    */
                err_ret = TRUE;
             }
             break;

          default:
             break;
      }
   }
   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

int chk_vmap (int c, int d)
/*                                       * tests for chromosome no., trait no.
 *                                         or interval no. out of range
 *                                         also tests interval proportions,
 *                                         external distances or external
 *                                         recombination fracts within range  */
{
   int   i, ix = 0;
   int   err_ret = FALSE;

   if (CH == c)                         /* checks chromosome no.              */
   {
      if (j_par[0] < 1 || j_par[0] > pmx->max_chrom)
      {
         err_typ = ERD_CHO;
         return FALSE;
      }
      ++ix;
   }
   if (j_par[ix] < 1 || j_par[ix] > pmx->max_trait)
   {                                    /* checks trait no.                   */
      err_typ = ERD_TRO;
      return FALSE;
   }
   ++ix;

   if (d == IM)                         /* proportions for intervals:         */
      for (i = 0; i < nk_par; ++i, ++ix) /* checks interval no's             */
         if (j_par[ix] < 1 || j_par[ix] > pmx->max_mark - 1)
         {
            err_typ = ERA_TIN;
            err_ret = TRUE;
         }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nx_par; i++)         /* checks mapping specs               */
   {
      switch (d)
      {
          case AM:
          case IM:                     /* proportions, one or all intervals:  */

             if (d_par[i] <= 0. || d_par[i] >= 1.)
             {
                err_typ = ERA_TIO;      /* requires:  0. < prop < 1.          */
                err_ret = TRUE;
             }
             break;

          case BRM:
          case ERM:
          case XRM:

             if (d_par[i] <= 0. || d_par[i] >= .5) /* external recomb fract: */
             {
                err_typ = ERA_TFO;      /* requires:  0. < rf < .5            */
                err_ret = TRUE;
             }
             break;

          case BDM:
          case EDM:
          case XDM:
          case KBDM:
          case KEDM:
          case KXDM:

             if (d_par[i] <= 0. || d_par[i] > 500.) /* external distance:    */
             {
                err_typ = ERA_TDO;      /* requires 0. < dist <= 500. cM      */
                err_ret = TRUE;
             }
             break;

          default:
             break;
      }
   }
   if (!err_ret)
      for (i = 1; i < nx_par; i++)
         if (d_par[i] <= d_par[i - 1])  /* checks for increasing order        */
         {
            if (d == IM || d == AM)
               err_typ = ERA_TID;
            else if (d == BRM || d == ERM || d == XRM)
               err_typ = ERA_TFD;
            else if (d == BDM || d == EDM || d == XDM
                     || d == KBDM || d == KEDM || d == KXDM)
               err_typ = ERA_TDD;
            err_ret = TRUE;
         }
   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

MAP  *chk_map_node (int c, int m, int sx)
/*                                       * tests for statement already given
 *                                         for this chromosome no. & gender
 *                                         returns map node, if found
 */
{
   MAP_PTR p_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   if (m == MARK_LOC)
      p_node = first_mark_map_node[sx];
   else
      p_node = first_trait_map_node[sx];

   if (p_node)
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERD_RMP;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

VMAP *chk_vmap_node (int c, int sx)
/*                                       * tests for any external or interval
 *                                         map statement already given for
 *                                         this chromosome & gender
 *                                         returns vmap node, if found
 */
{
   VMAP_PTR p_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   p_node = first_trait_vmap_node[sx];

   if (p_node)
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;       /* checks for vmap node               */

      if (p_node->chrom_no == n)
         return p_node;
   }
   return NULL;
}

int chk_vext (XMAP_PTR p_x, int c)
/*
 * checks for external or beginning or ending distances & recomb fractions
 */
{
   if (p_x)
   {
      if (CH == c)
         err_typ = ERA_OXC;
      else
         err_typ = ERA_OXP;
   }
   if (err_typ)
      return TRUE;
   else
      return FALSE;
}

IMAP_PTR chk_vint (VMAP_PTR p_v, int c, int d, int l)
/*                                       * checks for list of proportions
 *                                         already present for specified
 *                                         interval (0 = all)
 */
{
   IMAP_PTR p_i;

   p_i = p_v->p_intv;

   if (!p_i)
      return NULL;

   if (AM == d)                         /* checks for all intervals present   */
   {
      if (p_i->intv_no == 0)
      {
         if (CH == c)
            err_typ = ERA_OAC;
         else
            err_typ = ERA_OAP;
      }
      else
         p_i = NULL;
   }

   if (IM == d)                         /* checks for this interval present   */
   {
      while (p_i && p_i->p_next && p_i->intv_no < l)
         p_i = p_i->p_next;
      if (p_i && p_i->intv_no == l)
      {
         if (CH == c)
            err_typ = ERA_OIC;
         else
            err_typ = ERA_OIP;
      }
      else
         p_i = NULL;
   }

   return p_i;                          /* returns ptr to interval (or NULL)  */
}

int chk_locus (int c, int md)           /* tests for chromosome no., marker  */
                                        /* no's, trait no's, no. allele      */
                                        /* frequencies within range.         */
                                        /* checks sum of allele frequencies  */
{
   int   i, ix = 0;

   if (CH == c)
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_chrom)
      {
         err_typ = ERD_CHO;
         return FALSE;
      }
      ix++;
   }
   for (i = 0; i < nk_par; i++)
   {
      if (MARK_LOC == md)
      {
         if (j_par[ix] < 1 || j_par[ix] > pmx->max_mark)
         {
            err_typ = ERD_MKO;
            return FALSE;
         }
      }
      else if (j_par[ix] < 1 || j_par[ix] > pmx->max_trait)
      {
         err_typ = ERD_TRO;
         return FALSE;
      }
      ix++;
   }
   if (MARK_LOC == md)
   {
      if (nx_par < 2 || nx_par > pmx->max_allele)
      {
         err_typ = ERD_MLO;
         return FALSE;
      }
   }
   else
   {
      if (nx_par != 2)
      {
         err_typ = ERD_TLO;
         return FALSE;
      }
   }
   for (i = 0; i < nx_par; i++)
      if (d_par[i] < 0)
      {
         err_typ = ERD_AFO;
         return FALSE;
      }

   return TRUE;
}

LOCUS *chk_locus_node (int no, int c, int md) /* checks for allele freqs    */
                                        /* already given for this locus      */
{
   LOCUS_PTR p_node;
   int   n = 0;

   if (MARK_LOC == md)                  /* marker allele frequencies         */
   {
      if (CH == c)
         n = j_par[0];
      p_node = first_mark_locus_node;
   }
   else
   {                                    /* trait allele frequencies          */
      n = NO_CH_NO;
      p_node = first_trait_locus_node;
   }
   if (p_node)
   {
      while (p_node->p_next
             && (p_node->chrom_no < n
                 || (p_node->chrom_no == n && p_node->no < no)))
         p_node = p_node->p_next;
      if (p_node->chrom_no == n && p_node->no == no)
      {
         if (MARK_LOC == md)
            err_typ = ERD_MAF;          /* overrides previous allele freqs   */
         else                           /* .. for marker                     */
            err_typ = ERD_TAF;          /* .. or for trait                   */
         sprintf (errnumb, "%d", p_node->no);
         err_txt = errnumb;
         return p_node;
      }
   }
   return NULL;                         /* no match for id                   */
}
int chk_locus_norm (void)               /* tests for unnormalized freqs      */
{
   int   i;
   double sum = 0.;

   for (i = 0; i < nx_par; i++)         /* this test is done last            */
      sum += d_par[i];

   if (unnorm_frq (sum))
      err_typ = ERD_SMA;

   return TRUE;
}

int chk_means (void)                    /* tests for trait no. in range,     */
                                       /* and for no. means = 3              */
{
   int   i;

   for (i = 0; i < nk_par; i++)
      if (j_par[i] < 1 || j_par[i] > pmx->max_trait)
      {
         err_typ = ERD_TRO;
         return FALSE;
      }
   if (nx_par != 3)
   {
      err_typ = ERD_TGM;
      return FALSE;
   }
   return TRUE;
}

MEANS *chk_means_node (int no)          /* tests for trait genotype means    */
                                       /* already given for this trait       */
{
   MEANS_PTR p_node;

   p_node = first_means_node;
   if (p_node)
   {
      while (p_node->p_next && p_node->no < no)
         p_node = p_node->p_next;
      if (p_node->no == no)
      {
         err_typ = ERD_TGR;
         return p_node;
      }
   }
   return NULL;
}

int chk_vari (void)                     /* tests for trait no. in range,     */
{
   int   i;

   for (i = 0; i < nk_par; i++)
      if (j_par[i] < 1 || j_par[i] > pmx->max_trait)
      {
         err_typ = ERD_TRO;
         return FALSE;
      }
   if (d_par[0] < 0.)
   {
      err_typ = ERD_VMN;
      return FALSE;
   }
   return TRUE;
}

VARI *chk_vari_node (int no, int v)     /* tests for variance already given  */
                                       /* for this trait                     */
{
   VARI_PTR p_node;

   if (AV == v)
      p_node = first_addi_node;
   else
      p_node = first_resi_node;

   if (p_node)
   {
      while (p_node->p_next && p_node->no < no)
         p_node = p_node->p_next;

      if (p_node->no == no)
      {
         if (AV == v)
            err_typ = ERD_AVR;
         else
            err_typ = ERD_RVR;
         return p_node;
      }
   }
   return NULL;
}

int chk_mrkdat (int c)                  /* tests for chromosome no. out of   */
                                        /* range and for no. markers out of  */
                                        /* range                             */
                                        /* tests for ea pair of alleles > 0  */
                                        /* if observed, or 0 if unobserved   */
                                        /* checks for duplicate names        */
{
   int   i, j, k, ix = 0;
   int   err_ret = FALSE;

   if (CH == c)                         /* checks chromosome no.             */
   {
      if (j_par[0] < 1 || j_par[0] > pmx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (!err_ret)                        /* checks no. markers                */
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_mark)
      {
         err_typ = ERA_MNO;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   if (!err_ret)                        /* checks for alleles for each indiv */
   {                                    /* either unobserved, or all > 0     */
      j = ix + 1;
      for (i = 0; i < nm_par; ++i)
      {
         for (k = 0; k < j_par[ix]; ++k, j += 2)
         {
            if (j_par[j] < 0 || j_par[j + 1] < 0)
            {
               err_typ = ERA_PAL;
               sprintf (errnumb, "%s", m_par[i]);
               err_txt = errnumb;
               err_ret = TRUE;
            }
            else if (!j_par[j] != !j_par[j + 1])
            {
               err_typ = ERA_UAL;
               sprintf (errnumb, "%s", m_par[i]);
               err_txt = errnumb;
               err_ret = TRUE;
            }
         }
      }
   }
   if (!err_ret)                        /* checks for duplicate names        */
      for (i = 1; i < nm_par; ++i)
         for (j = 0; j < i; ++j)
            if (!(strncmp (m_par[i], m_par[j], MAXNAMLEN + 1)))
            {
               err_typ = ERA_DIN;
               sprintf (errnumb, "%s", m_par[j]);
               err_txt = errnumb;
               err_ret = TRUE;
            }

   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

MARKDAT *chk_mrkdat_node (int c)        /* tests for statement already given */
                                        /* for this chromosome no.           */
{
   MARKDAT_PTR p_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   if ((p_node = first_markdat_node))
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERA_RMD;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

int chk_mrknam (int c)                  /* tests for chromosome no. out of   */
                                        /* range and for no. markers out of  */
                                        /* range                             */
                                        /* checks for duplicate names        */
{
   int   i, j, ix = 0;
   int   err_ret = FALSE;

   if (CH == c)                         /* checks chromosome no.             */
   {
      if (j_par[0] < 1 || j_par[0] > pmx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (!err_ret)                        /* checks no. markers                */
      if (nm_par < 1 || nm_par > pmx->max_mark)
      {
         err_typ = ERA_MNO;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   if (!err_ret)                        /* checks for duplicate names        */
      for (i = 1; i < nm_par; ++i)
         for (j = 0; j < i; ++j)
            if (!(strncmp (m_par[i], m_par[j], MAXNAMLEN + 1)))
            {
               err_typ = ERA_DMN;
               sprintf (errnumb, "%s", m_par[j]);
               err_txt = errnumb;
               err_ret = TRUE;
            }

   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

MARKNAM *chk_mrknam_node (int c)        /* tests for statement already given */
                                        /* for this chromosome no.           */
{
   MARKNAM_PTR p_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   if ((p_node = first_marknam_node))
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERA_RMN;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

int chk_prbgam (int cp, int ss)
/*                                       * tests for component no. out of
 *                                         range
 *                                         checks name & gamete (proband
 *                                         gametes cannot be duplicated
 *                                         within the set, & gametes must be
 *                                         0 (maternal) or 1 (paternal))      */
{
   int   i, j, ix = 0;
   int   err_ret = FALSE;

   if (CMP == cp)                       /* checks component no.               */
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_n_cmp)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   if (SC == ss)                        /* checks scoreset no.                */
   {
      if (j_par[ix] < 1)
      {
         err_typ = ERA_SCS;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   if (nm_par < 2)                      /* checks min no. proband gametes     */
   {
      err_typ = ERA_PGL;
      err_ret = TRUE;
   }
   else if (nm_par > MAXGAM)            /* checks max no. proband gametes     */
   {
      err_typ = ERA_PGX;
      sprintf (errnumb, "%d", MAXGAM);
      err_txt = errnumb;
      err_ret = TRUE;
   }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nm_par; ++i)
   {
      if (j_par[ix + i] != 0 && j_par[ix + i] != 1) /* checks haplotypes     */
      {
         err_typ = ERA_BGM;
         sprintf (errnumb, "%d", j_par[ix + i]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      for (j = 0; j < i; ++j)
         if (!(strcmp (m_par[i], m_par[j])) && j_par[ix + i] == j_par[ix + j])
         {                              /* checks for dup proband gametes  */
            err_typ = ERA_DPG;
            sprintf (errnumb, "%s %d", m_par[j], j_par[ix + i]);
            err_txt = errnumb;
            err_ret = TRUE;
         }
   }
   if (err_ret)
      return FALSE;

   return TRUE;
}

PROBGAM *chk_prbgam_node (int cp, int ss) /* tests for statement already    */
{                                       /* given for this comp and scoreset   */
   PROBGAM_PTR p_node;

   int   c, s, ix = 0;

   if (CMP == cp)
      c = j_par[ix++];
   else
      c = 1;
   if (SC == ss)
      s = j_par[ix++];
   else
      s = 1;

   if ((p_node = first_probgam_node))
   {
      while (p_node->p_next &&
             (p_node->component < c
              || (p_node->component == c && p_node->scoreset < s)))
         p_node = p_node->p_next;

      if (p_node->component == c && p_node->scoreset == s)
      {
         err_typ = ERA_OSC;
         return p_node;
      }
   }
   return NULL;
}

int chk_indnam (int cp)                 /* tests for component no. out of */
                                        /* range                              */
                                        /* checks names for duplication       */
{
   int   i, j;
   int   err_ret = FALSE;

   if (CMP == cp)                       /* checks component no.               */
   {
      if (j_par[0] < 1 || j_par[0] > pmx->max_n_cmp)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[0]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nm_par; ++i)
   {
      for (j = 0; j < i; ++j)
         if (!(strcmp (m_par[i], m_par[j])))
         {                              /* checks for dup named individuals  */
            err_typ = ERA_DAI;
            sprintf (errnumb, "%s", m_par[j]);
            err_txt = errnumb;
            err_ret = TRUE;
         }
   }
   if (err_ret)
      return FALSE;

   return TRUE;
}

INDNAM *chk_indnam_node (int cp, int nt) /* tests for statement already given */
{                                       /* for this component                 */
   INDNAM_PTR p_node;

   int   c, ix = 0;

   if (CMP == cp)
      c = j_par[ix++];
   else
      c = 1;

   if (AFFIND == nt)
      p_node = first_affnam_node;
   else if (DISPIND == nt)
      p_node = first_dispnam_node;
   if (p_node)
   {
      while (p_node->p_next && p_node->component < c)
         p_node = p_node->p_next;

      if (p_node->component == c)
      {
         err_typ = ERA_OAI;
         return p_node;
      }
   }
   return NULL;
}

int chk_wind (int c)                    /* tests for chromosome no. or    */
{                                       /* locus window size out of range     */
   int   ix = 0;

   if (CH == c)
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_chrom)
      {
         err_typ = ERD_CHO;
         return FALSE;
      }
      ix++;
   }
   if (j_par[ix] < 2)
   {
      err_typ = ERR_MNW;
      return FALSE;
   }
   else if (j_par[ix] > pmx->max_mark + 1)
   {
      err_typ = ERR_MXW;
      return FALSE;
   }

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

LOCWND *chk_wind_node (int c)           /* tests for statement already given */
{                                       /* for this chromosome no.            */
   LOCWND_PTR p_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   if ((p_node = first_locwind_node))
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERA_RLW;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

int chk_wndpat (int cp, int ss)         /* tests for component no., scoreset */
                                        /* no. and no. of patterns out of     */
                                        /* range ???                          */
                                        /* tests for pattern label > 0        */
{
   int   i, ix = 0;
   int   err_ret = FALSE;

   if (CMP == cp)                       /* checks component no.               */
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_chrom)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   if (SC == ss)                        /* checks scoreset no.                */
   {
      if (j_par[ix] < 1)
      {
         err_typ = ERA_SCS;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nk_par; ++i, ++ix)   /* checks pattern labels out of range */
   {
      if (j_par[ix] < 0 || j_par[ix] > CDgg[MAXGAM])
      {
         err_typ = ERA_WPO;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   }
   if (err_ret)
      return FALSE;

   return TRUE;
}

WNDPAT *chk_wndpat_node (int cp, int ss) /* tests for statement already    */
{                                       /* given for this comp and scoreset   */
   WNDPAT_PTR p_node;

   int   c, s, ix = 0;

   if (CMP == cp)
      c = j_par[ix++];
   else
      c = 1;
   if (SC == ss)
      s = j_par[ix++];
   else
      s = 1;

   if ((p_node = first_windpat_node))
   {
      while (p_node->p_next &&
             (p_node->component < c
              || (p_node->component == c && p_node->scoreset < s)))
         p_node = p_node->p_next;

      if (p_node->component == c && p_node->scoreset == s)
      {
         err_typ = ERA_OSP;
         return p_node;
      }
   }
   return NULL;
}

int chk_kinshp (int cmp, int m)         /* tests for component out of range  */
{                                       /* checks for pairs of names          */
   int   ix = 0;

   if (CMP == cmp)
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_n_cmp)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
      }
      ix++;
   }
   if (err_typ)
      return FALSE;

   if (nm_par % m)
   {
      sprintf (errnumb, "pairs");
      err_txt = errnumb;
      err_typ = ERK_KNN;
   }

   if (err_typ)
      return FALSE;
   return TRUE;
}

KINS *chk_kinshp_node (int cmp, int kc) /* tests for statement already        */
{                                       /* given for this component           */
   KINS *p_node;

   int   n = 0;

   if (CMP == cmp)
      n = j_par[0];

   if ((kc == KINSHP && (p_node = first_kins_node)) /* determines node type  */
       || (kc == INBRD && (p_node = first_inbreed_node))
       || (kc == INBRD2 && (p_node = first_2loc_node)))
   {
      while (p_node->p_next && p_node->component < n)
         p_node = p_node->p_next;

      if (p_node->component == n)
      {
         if (n == 0)
            err_typ = ERR_RED;
         else
            err_typ = ERK_RCP;
         return p_node;
      }
   }
   return NULL;
}

int chk_penetr (void)
{
   int   i;

   for (i = 0; i < nx_par; ++i)
      if (d_par[i] < 0. || d_par[i] > 1.)
         err_typ = ERA_RPV;
      else if (d_par[1] > d_par[2] || d_par[0] > d_par[1])
         err_typ = ERA_OPV;

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

int chk_reco_frq (void)
{
   int   i;

   for (i = 0; i < nx_par; ++i)
      if (d_par[i] < 0.)
         err_typ = ERK_NRF;
      else if (d_par[i] > .5)
         err_typ = ERK_XRF;

   if (err_typ)
      return FALSE;
   return TRUE;
}

int chk_pseudo (void)
{
   int   i;

/*
   for (i = 0; i < nx_par; ++i)
   if (d_par[i] == 0)
   {
   d_par[i] = DIFF;
   err_typ = ERA_PPW;
   }
 */
   for (i = 0; i < nx_par; ++i)
      if (d_par[i] < 0)
         err_typ = ERA_PPN;

   if (err_typ)
      return FALSE;
   return TRUE;
}

void chk_seeds (void)
{
   if (u_par[0] % 2 == 0)
   {
      u_par[0] |= 1;
      err_typ = ERD_SEE;
   }
   if (u_par[1] == 0)
   {
      u_par[1] = 1;
      err_typ = ERD_SEE;
   }
   return;
}

int unnorm_frq (double q_sum)
{
   if (q_sum < Q_SUMMN || q_sum > Q_SUMMX) /* checks unnormlzd */
      return TRUE;
   else
      return FALSE;
}

int unnorm_frq_mx (int norm_flg, double q_sum)
{
   if (!norm_flg && q_sum > Q_SUMMX)    /* checks sum > upper limit           */
      return TRUE;
   else
      return FALSE;
}

/* ---------------- free fncs:  frees list nodes or arrays ----------------- */

void free_unlnkd (void)                 /* frees current list of unlinked */
                                        /* traits                            */
{
   UNLNKD_PTR p_node, p_nxtn;

   p_node = first_unlnkd_node;

   while (p_node != NULL)
   {
      p_nxtn = p_node->p_next;
      free (p_node);
      p_node = p_nxtn;
   }
   first_unlnkd_node = NULL;

   return;
}

void free_mark_list (SIM_PTR p_node)    /* free list of markers to select     */
{
   if (p_node && p_node->p_mark)
   {
      free (p_node->p_mark);
      p_node->p_mark = NULL;
      p_node->no_mark_sel = 0;
   }
   return;
}

void free_dist (MAP_PTR p_node)         /* frees list of map distances        */
{
   if (p_node && p_node->p_dist)
   {
      free (p_node->p_dist);
      p_node->p_dist = NULL;
      p_node->no_loci = 0;
   }
   return;
}

void free_vdist (XMAP_PTR p_x)          /* frees list of external dists       */
{
   if (p_x && p_x->no_pos)
   {
      free (p_x->p_dist);
      p_x->p_dist = NULL;
      free (p_x->p_rf);
      p_x->p_rf = NULL;
      p_x->no_pos = 0;
   }
   return;
}

void free_vprop (IMAP_PTR p_i)          /* frees list of interval proportions */
{

   if (p_i && p_i->p_prop)
      free (p_i->p_prop);
   p_i->p_prop = NULL;
   p_i->no_prop = nx_par;

   return;
}

void free_freq (LOCUS_PTR p_locus)      /* if there are no loci for the      */
                                        /* frequencies, frees list           */
{
   if ((p_locus->p_freq_node->no_loci -= 1) == 0)
   {                                    /* decrement & test no. loci         */
      free (p_locus->p_freq_node->p_freq);
      free (p_locus->p_freq_node);
   }
   p_locus->p_freq_node = NULL;
   p_locus->no_alleles = 0;

   return;
}

void free_mrkdat_list (MARKDAT_PTR p_node)
{
   if (p_node->p_obs)
   {
      free (p_node->p_obs);
      p_node->p_obs = NULL;
   }
   if (p_node->p_alls != NULL)
   {
      free (p_node->p_alls);
      p_node->p_alls = NULL;
   }
   return;
}

void free_mrknam_list (MARKNAM_PTR p_node)
{
   if (p_node->p_nams)
   {
      free (p_node->p_nams);
      p_node->p_nams = NULL;
   }
   return;
}

void free_prbgam_list (PROBGAM_PTR p_node)
{
   if (p_node->p_pglist)
   {
      free (p_node->p_pglist);
      p_node->p_pglist = NULL;
   }
   p_node->no_pgams = 0;

   return;
}

void free_indnam_list (INDNAM_PTR p_node)
{
   if (p_node->p_namlist)
   {
      free (p_node->p_namlist);
      p_node->p_namlist = NULL;
   }
   p_node->no_nams = 0;

   return;
}

void free_wndpat_set (WNDPAT_PTR p_node)
{
   if (p_node->p_pats)
   {
      free (p_node->p_pats);
      p_node->p_pats = NULL;
   }
   p_node->no_pats = 0;

   return;
}

void free_nams_list (KINS * p_node)
{
   if (p_node->p_nams != NULL)
   {
      free (p_node->p_nams);
      p_node->p_nams = NULL;
   }
   return;
}

void free_reco_frq (double **r)
{
   if (*r)
   {
      free (*r);
      *r = NULL;
   }
   return;
}

/* --------------- alloc fncs:  malloc list nodes or arrays ---------------- */

SIM  *alloc_sim_node (int c)            /* allocates the sim node             */
{                                       /* inserts in list in numeric order   */
   SIM_PTR new_node, cur_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   new_node = malloc (sizeof (SIM));
   alloc_chk (WHERE, new_node);

   if (!first_sim_node || first_sim_node->chrom_no > n)
   {
      new_node->p_next = first_sim_node; /* puts at start of list            */
      first_sim_node = new_node;        /* (may be only node)                 */
   }
   else
   {
      cur_node = first_sim_node;        /* inserts node in list               */
      /*                                   or adds to end of list             */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_mark_list (SIM_PTR p_node)   /* allocates the marker list                  */
{
   p_node->p_mark = malloc (p_node->no_mark_sel * sizeof (int));
   alloc_chk (WHERE, p_node->p_mark);

   return;
}

UNLNKD *alloc_unlnkd_node (int no)      /* allocates an unlinked trait node   */
{
   UNLNKD_PTR new_node, cur_node;

   new_node = malloc (sizeof (UNLNKD));
   alloc_chk (WHERE, new_node);

   if (!first_unlnkd_node || first_unlnkd_node->no > no)
   {
      new_node->p_next = first_unlnkd_node; /* puts at start of list         */
      first_unlnkd_node = new_node;
   }
   else
   {
      cur_node = first_unlnkd_node;     /* puts after start of list           */
      while (cur_node->p_next && cur_node->p_next->no < no)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

TRTPED *alloc_trtped_node (int no)      /* allocates the traitped node        */
                                         /* inserts in list in numeric order  */
{
   TRTPED_PTR new_node, cur_node;

   new_node = malloc (sizeof (TRTPED));
   alloc_chk (WHERE, new_node);

   if (!first_trtped_node || first_trtped_node->no > no)
   {
      new_node->p_next = first_trtped_node;
      first_trtped_node = new_node;
   }
   else
   {
      cur_node = first_trtped_node;
      while (cur_node->p_next && cur_node->p_next->no < no)
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

MAP  *add_map_node (int c, int m, int sx) /* allocates the map node          */
                                        /* inserts in list in numeric order   */
{
   MAP_PTR new_node, cur_node;
   MAP_PTR *first_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   new_node = malloc (sizeof (MAP));
   alloc_chk (WHERE, new_node);

   if (m == MARK_LOC)                   /* checks for marker or trait map     */
      first_node = &first_mark_map_node[sx];
   else
      first_node = &first_trait_map_node[sx];

   if (!*first_node || (*first_node)->chrom_no > n)
   {
      new_node->p_next = *first_node;   /* puts at start of list              */
      *first_node = new_node;           /* (may be only node)                 */
   }
   else
   {                                    /* inserts node in list or adds       */
      cur_node = *first_node;           /* to end of list                     */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   new_node->chrom_no = n;

   return new_node;
}

VMAP *add_vmap_node (int c, int sx)     /* allocates the map node             */
                                        /* inserts in list in chromosome      */
                                        /* and trait order                    */
{
   VMAP_PTR new_node, cur_node, first_node;

   int   n = 0;
   int   ix = 0;

   if (CH == c)
      n = j_par[ix++];

   new_node = malloc (sizeof (VMAP));
   alloc_chk (WHERE, new_node);

   first_node = first_trait_vmap_node[sx];

   if (!first_node || first_node->chrom_no > n)
   {
      new_node->p_next = first_node;    /* puts at start of list              */
      first_trait_vmap_node[sx] = new_node; /* (may be only node)             */
   }
   else
   {
      cur_node = first_node;            /* inserts node in list or at end     */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   new_node->chrom_no = n;
   new_node->no = j_par[ix];
   new_node->p_beg = NULL;
   new_node->p_end = NULL;
   new_node->p_intv = NULL;

   return new_node;
}

void alloc_dist (MAP_PTR p_node, int m) /* allocates the array             */
{
   if (m == MARK_LOC)
      p_node->p_dist = malloc ((nx_par + 1) * sizeof (DIST));
   else
      p_node->p_dist = malloc (nx_par * sizeof (DIST));
   alloc_chk (WHERE, p_node->p_dist);

   return;
}

XMAP_PTR alloc_vext (void)
/*                                       * allocates the beg or ending map    */
{
   XMAP_PTR p;

   p = malloc (sizeof (XMAP));
   alloc_chk (WHERE, p);

   return p;
}

IMAP_PTR add_vint (VMAP_PTR p_v, int l)
/*                                       * adds  the interval prop list       */
{
   IMAP_PTR p_new, p_i;

   p_new = malloc (sizeof (IMAP));
   alloc_chk (WHERE, p_new);

   p_i = p_v->p_intv;

   if (!p_i)
   {
      p_v->p_intv = p_new;              /* adds as only interval              */
      p_new->p_next = NULL;
   }
   else if (p_i->intv_no > l)
   {
      p_new->p_next = p_v->p_intv;      /* puts at start of list of intervals */
      p_v->p_intv = p_new;
   }
   else
   {
      while (p_i->p_next && p_i->p_next->intv_no < l)
         p_i = p_i->p_next;
      p_new->p_next = p_i->p_next;      /* adds to list of intervals          */
      p_i->p_next = p_new;
   }
   p_new->intv_no = l;
   p_new->no_prop = nx_par;

   return p_new;
}

void alloc_vdist (XMAP_PTR p_x)
/*                                       * allocates extern dist & rf lists   */
{
   p_x->p_dist = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, p_x->p_dist);
   p_x->p_rf = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, p_x->p_rf);

   return;
}

void alloc_vprop (IMAP_PTR p_i)
/*                                       * allocates internal prop list       */
{
   p_i->p_prop = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, p_i->p_prop);

   return;
}

FREQ_PTR alloc_frq (void)               /* allocates frequency header        */
{                                       /* and array                         */
   return (alloc_frqx (nx_par));
}

FREQ_PTR alloc_dflt_frq (void)          /* allocates frequency header        */
{                                       /* and array for default values      */
   return (alloc_frqx (2));
}

FREQ_PTR alloc_frqx (int n_alleles)     /* allocates frequency header        */
{                                       /* and array                         */
   FREQ_PTR p_f;

   p_f = malloc (sizeof (FREQ));
   alloc_chk (WHERE, p_f);
   p_f->p_freq = malloc (n_alleles * sizeof (double));
   alloc_chk (WHERE, p_f->p_freq);
   return p_f;
}

LOCUS_PTR alloc_locus_node (int no, int c, int md)
{
   int   cno = 0;

   if (MARK_LOC == md)                  /* marker allele frequencies         */
   {
      if (CH == c)
         cno = j_par[0];
   }
   else                                 /* trait allele frequencies          */
      cno = NO_CH_NO;

   return (alloc_locus_nodex (cno, no, md));
}

LOCUS_PTR alloc_dflt_locus_node (int cno, int no, int md)
{
   return (alloc_locus_nodex (cno, no, md));
}

LOCUS *alloc_locus_nodex (int chrom, int no, int md)
{
   LOCUS_PTR new_node, cur_node;
   LOCUS_PTR *first_node;

   if (MARK_LOC == md)
      first_node = &first_mark_locus_node;
   else
      first_node = &first_trait_locus_node;

   new_node = malloc (sizeof (LOCUS));
   alloc_chk (WHERE, new_node);

   if (!*first_node
       || (*first_node)->chrom_no > chrom
       || ((*first_node)->chrom_no == chrom && (*first_node)->no > no))
   {
      new_node->p_next = *first_node;   /* puts at start of list             */
      *first_node = new_node;
   }
   else
   {
      cur_node = *first_node;

      while (cur_node->p_next
             && (cur_node->p_next->chrom_no < chrom
                 || (cur_node->p_next->chrom_no == chrom
                     && cur_node->p_next->no < no)))
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

MEANS *alloc_means_node (int no)        /* allocates the means node          */
                                         /* inserts in list in numeric order */
{
   MEANS_PTR new_node, cur_node;

   new_node = malloc (sizeof (MEANS));
   alloc_chk (WHERE, new_node);

   if (!first_means_node || first_means_node->no > no)
   {
      new_node->p_next = first_means_node;
      first_means_node = new_node;
   }
   else
   {
      cur_node = first_means_node;
      while (cur_node->p_next && cur_node->p_next->no < no)
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

VARI *alloc_vari_node (int no, int v)   /* allocates the variance node */
                                         /* inserts in list in numeric  */
                                         /* trait order                 */
{
   VARI_PTR new_node, cur_node;
   VARI_PTR *first_node;

   if (AV == v)
      first_node = &first_addi_node;
   else
      first_node = &first_resi_node;

   new_node = malloc (sizeof (VARI));
   alloc_chk (WHERE, new_node);

   if (!*first_node || (*first_node)->no > no)
   {
      new_node->p_next = *first_node;
      *first_node = new_node;
   }
   else
   {
      cur_node = *first_node;
      while (cur_node->p_next && cur_node->p_next->no < no)
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

MARKDAT *alloc_mrkdat_node (int c)      /* allocates marker data node        */
{                                       /* inserts in list in numeric  order */
   MARKDAT_PTR new_node, cur_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   new_node = malloc (sizeof (MARKDAT));
   alloc_chk (WHERE, new_node);

   if (!first_markdat_node || first_markdat_node->chrom_no > n)
   {
      new_node->p_next = first_markdat_node; /* puts at start of list       */
      first_markdat_node = new_node;    /* (may be only node)                */
   }
   else
   {                                    /* inserts node in list or           */
      cur_node = first_markdat_node;    /* adds to end of list               */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_mrkdat_list (MARKDAT_PTR p_node) /* allocates marker data array */
{
   p_node->p_obs = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_obs);

   p_node->p_alls = malloc (nk_par * sizeof (int));
   alloc_chk (WHERE, p_node->p_alls);

   return;
}

MARKNAM *alloc_mrknam_node (int c)      /* allocates marker name node        */
{                                       /* inserts in list in numeric  order */
   MARKNAM_PTR new_node, cur_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   new_node = malloc (sizeof (MARKNAM));
   alloc_chk (WHERE, new_node);

   if (!first_marknam_node || first_marknam_node->chrom_no > n)
   {
      new_node->p_next = first_marknam_node; /* puts at start of list       */
      first_marknam_node = new_node;    /* (may be only node)                */
   }
   else
   {                                    /* inserts node in list or           */
      cur_node = first_marknam_node;    /* adds to end of list               */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

                                              /* FIXME:  generic name fnc    */
void alloc_mrknam_list (MARKNAM_PTR p_node) /* allocates marker name array */
{
   p_node->p_nams = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_nams);

   return;
}

PROBGAM *alloc_prbgam_node (int cp, int ss) /* allocates proband gamete node  */
                                        /* inserts in list in order of        */
                                        /* component and scoreset             */
{
   PROBGAM_PTR new_node, cur_node;

   int   c, s, ix = 0;

   if (CMP == cp)
      c = j_par[ix++];
   else
      c = 1;
   if (SC == ss)
      s = j_par[ix++];
   else
      s = 1;

   new_node = malloc (sizeof (PROBGAM));
   alloc_chk (WHERE, new_node);

   if (!first_probgam_node || first_probgam_node->component > c
       || (first_probgam_node->component == c
           && first_probgam_node->scoreset > s))
   {
      new_node->p_next = first_probgam_node; /* puts at start of list        */
      first_probgam_node = new_node;    /* (may be only node)                 */
   }
   else
   {
      cur_node = first_probgam_node;    /* inserts node in list               */
      while (cur_node->p_next           /* or adds to end of list             */
             && (cur_node->p_next->component < c
                 || (cur_node->p_next->component == c
                     && cur_node->p_next->scoreset < s)))
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_prbgam_list (PROBGAM_PTR p_node) /* allocates the proband        */
{                                       /* gamete array                       */

   p_node->p_pglist = malloc ((nm_par) * sizeof (PROBANDS));
   alloc_chk (WHERE, p_node->p_pglist);

   return;
}

INDNAM *alloc_indnam_node (int cp, int nt) /* allocates named individs node   */
                                        /* (affected or display individuals)  */
                                        /* inserts in list in order of compnt */
{
   INDNAM_PTR *pp_first;
   INDNAM_PTR new_node, cur_node;

   int   c;

   if (CMP == cp)
      c = j_par[0];
   else
      c = 1;

   new_node = malloc (sizeof (INDNAM));
   alloc_chk (WHERE, new_node);

   if (AFFIND == nt)
      pp_first = &first_affnam_node;
   else if (DISPIND == nt)
      pp_first = &first_dispnam_node;

   if (!*pp_first || (*pp_first)->component > c)
   {
      new_node->p_next = *pp_first;     /* puts at start of list              */
      *pp_first = new_node;             /* (may be only node)                 */
   }
   else
   {
      cur_node = *pp_first;             /* inserts node in list               */
      while (cur_node->p_next           /* or adds to end of list             */
             && cur_node->p_next->component < c)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_indnam_list (INDNAM_PTR p_node) /* allocates the named individuals */
{                                       /* list                               */

   p_node->p_namlist = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_namlist);

   return;
}

LOCWND *alloc_wind_node (int c)         /* allocates window locus node    */
{                                       /* inserts in list in numeric order   */
   LOCWND_PTR new_node, cur_node;

   int   n = 0;

   if (CH == c)
      n = j_par[0];

   new_node = malloc (sizeof (LOCWND));
   alloc_chk (WHERE, new_node);

   if (!first_locwind_node || first_locwind_node->chrom_no > n)
   {
      new_node->p_next = first_locwind_node; /* puts at start of list         */
      first_locwind_node = new_node;    /* (may be only node)                 */
   }
   else
   {
      cur_node = first_locwind_node;    /* inserts node in list               */
      /*                                 * or adds to end of list             */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

WNDPAT *alloc_wndpat_node (int cp, int ss) /* allocates window pattern node   */
{                                       /* inserts in list in order of        */
   WNDPAT_PTR new_node, cur_node;       /* component and scoreset             */

   int   c, s, ix = 0;

   if (CMP == cp)
      c = j_par[ix++];
   else
      c = 1;
   if (SC == ss)
      s = j_par[ix++];
   else
      s = 1;

   new_node = malloc (sizeof (WNDPAT));
   alloc_chk (WHERE, new_node);

   if (!first_windpat_node || first_windpat_node->component > c
       || (first_windpat_node->component == c
           && first_windpat_node->scoreset > s))
   {
      new_node->p_next = first_windpat_node; /* puts at start of list         */
      first_windpat_node = new_node;    /* (may be only node)                 */
   }
   else
   {
      cur_node = first_windpat_node;    /* inserts node in list               */
      while (cur_node->p_next           /* or adds to end of list             */
             && (cur_node->p_next->component < c
                 || (cur_node->p_next->component == c
                     && cur_node->p_next->scoreset < s)))
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_wndpat_set (WNDPAT_PTR p_node) /* allocates window pattern set     */
{
   p_node->p_pats = malloc ((nk_par) * sizeof (int));
   alloc_chk (WHERE, p_node->p_pats);

   return;
}

KINS *alloc_kinshp_node (int cmp, int kc) /* allocates kinship node.          */
                                        /* inserts in list after last set     */
                                        /* for the chromosome                 */
{
   KINS *new_node, *cur_node;
   KINS **p_first;
   int   cp = 1;

   if (CMP == cmp)
      cp = j_par[0];

   new_node = malloc (sizeof (KINS));
   alloc_chk (WHERE, new_node);

   if (kc == KINSHP)
      p_first = &first_kins_node;
   else if (kc == INBRD)
      p_first = &first_inbreed_node;
   else if (kc == INBRD2)
      p_first = &first_2loc_node;

   if (!*p_first || (*p_first)->component > cp)
   {
      new_node->p_next = *p_first;      /* puts at start of list             */
      *p_first = new_node;              /* (may be only node)                */
   }
   else
   {                                    /* inserts node in list              */
      cur_node = *p_first;              /* or adds to end of list            */
      while (cur_node->p_next && cur_node->p_next->component <= cp)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_nams_list (KINS * p_node)    /* allocates the list of pedigree    */
{                                       /* member names                      */

   p_node->p_nams = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_nams);

   return;
}

double *alloc_reco_frq (void)           /* allocates list of recomb freqs    */
{
   double *r;

   r = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, r);

   return r;
}

/* -------------- stor fncs:  stores data in nodes or arrays ---------------- */

void stor_sim_data (SIM_PTR p_node, int c, int m, int t) /* stores the marker */
                                        /* specification type, chromosome no. */
                                        /* the no. markers, a pointer to the  */
                                        /* list of markers (for the Autozyg   */
                                        /* progs) and the no. traits in the   */
                                        /* simulation node                    */
{
   int   ix = 0;

   p_node->sim_typ = m;

   if (CH == c)
      p_node->chrom_no = j_par[ix++];
   else
      p_node->chrom_no = 0;

   if (MK == m)                         /* sets no. markers to simulate       */
      p_node->no_mark_sel = j_par[ix++];
   else if (AMK == m)                   /* or to select all markers           */
      p_node->no_mark_sel = 0;
   else if (MKL == m)                   /* or to select a list of markers     */
   {
      p_node->no_mark_sel = nk_par;
      ix += nk_par;
   }

   p_node->no_mark_map = 0;             /* reserves slot for no. markers mapd */

   if (TR == t)                         /* sets no. traits to simulate/select */
      p_node->no_traits = j_par[ix];
   else if (TRL == t)                   /* or trait locus to simulate/select  */
      p_node->no_traits = 1;
   else
      p_node->no_traits = 0;

   p_node->p_mark = NULL;

   return;
}

void stor_mark_list (SIM_PTR p_node, int c) /* stores the list of selected  */
                                              /* markers, sorted by mark. no. */
{
   int   i;
   int  *p_m;

   i = CH == c ? 1 : 0;

   for (p_m = p_node->p_mark; p_m < p_node->p_mark + p_node->no_mark_sel; ++p_m)
      *p_m = j_par[i++];

   qsort (p_node->p_mark, p_node->no_mark_sel, sizeof (int), qfunc_int);

   return;
}

void stor_unlnkd (int no, UNLNKD_PTR p_node) /* stores the trait no.        */
                                        /* in the unlinked node               */
{
   p_node->no = no;

   return;
}

void stor_trtped_pos (int typ, int no, int *pos, TRTPED_PTR p_node)
{
   int   m;

   m = typ == INTDAT2 ? 3 : 2;

   p_node->no = no;
   p_node->dat_typ = typ;
   p_node->pos[0] = pos[0];
   if (typ == INTDAT2)
      p_node->pos[1] = pos[1];

   return;
}

void stor_map (MAP_PTR p_node, int c, int d, int m) /* stores no. of loci.  */
                                       /* stores an array of trait or marker  */
                                       /* id's, preceding marker id's,        */
                                       /* probabilities & distances in the    */
                                       /* distance structure.                 */
                                       /* for traits, sorts the array         */
{
   int   i;
   int   ix = 0, jx = 0;
   DIST_PTR p_r;

   if (CH == c)
      ++ix;

   if (m == MARK_LOC)
   {
      if (PM != d && KPM != d)
         p_node->no_loci = nx_par + 1;
      else
         p_node->no_loci = nx_par;
   }
   else
      p_node->no_loci = nx_par;

   p_r = p_node->p_dist;
   for (i = 0; i < p_node->no_loci; ++i)
   {
      if (m == MARK_LOC)
      {
         p_r->no = i + 1;
         p_r->map_mark = i;
      }
      else
      {
         p_r->no = j_par[ix];
         p_r->map_mark = j_par[ix + p_node->no_loci];
         ++ix;
      }

      if (i == 0 && m == MARK_LOC)
      {
         p_r->dist = 0.;
         p_r->prob = .5;
      }
      else
      {
         switch (d)
         {
             case RM:
                p_r->dist = -.5 * log (1. - 2. * d_par[jx]) * 100.;
                p_r->prob = d_par[jx];
                break;
             case DM:
                p_r->dist = d_par[jx];
                p_r->prob = .5 * (1. - exp (-2. * d_par[jx] / 100.));
                break;
             case PM:
                p_r->dist = d_par[jx + 1] - d_par[jx];
                p_r->prob = .5 *
                   (1. - exp (-2. * (d_par[jx + 1] - d_par[jx]) / 100.));
                break;
             case KDM:
                p_r->prob = .5 *
                   (((exp (4. * d_par[jx] / 100.)) - 1.)
                    / ((exp (4. * d_par[jx] / 100.)) + 1.));
                p_r->dist = -.5 * log (1. - 2. * p_r->prob) * 100.;
                break;
             case KPM:
                p_r->dist = d_par[jx + 1] - d_par[jx];
                p_r->prob = .5 *
                   (((exp (4. * p_r->dist / 100.)) - 1.)
                    / ((exp (4. * p_r->dist / 100.)) + 1.));
                p_r->dist = -.5 * log (1. - 2. * p_r->prob) * 100.;
                break;
             default:
                break;
         }
         ++jx;
      }
      ++p_r;
   }
   if (m == TRAIT_LOC)
      qsort (p_node->p_dist, p_node->no_loci, sizeof (DIST), compar_mkid_d);

   return;
}

void stor_vdist (XMAP_PTR p_x, int d, int qsf ()) /* stores the list of    */
                                        /* external dists and recomb fracts   */
{
   int   i;

   p_x->no_pos = nx_par;

   qsort (d_par, nx_par, sizeof (double), qsf); /* sorts: incr or decr order */

   if (XDM == d || BDM == d || EDM == d)
      for (i = 0; i < nx_par; ++i)
      {
         p_x->p_dist[i] = d_par[i];
         p_x->p_rf[i] = dist2rf (d_par[i]);
      }
   else if (XRM == d || BRM == d || ERM == d)
      for (i = 0; i < nx_par; ++i)
      {
         p_x->p_dist[i] = rf2dist (d_par[i]);
         p_x->p_rf[i] = d_par[i];
      }
   if (KXDM == d || KBDM == d || KEDM == d)
      for (i = 0; i < nx_par; ++i)
      {
         p_x->p_rf[i] = .5 *
            (((exp (4. * d_par[i] / 100.)) - 1.)
             / ((exp (4. * d_par[i] / 100.)) + 1.));
         p_x->p_dist[i] = -.5 * log (1. - 2. * p_x->p_rf[i]) * 100.;
      }
   return;
}

void stor_vprop (IMAP_PTR p_i)          /* stores the list of interval        */
                                        /* proportions for the specified intv */
{
   int   i;

/* qsort (d_par, nx_par, sizeof (double), qsf); ***                           */

   for (i = 0; i < nx_par; ++i)
      p_i->p_prop[i] = d_par[i];

   return;
}

int compar_mkid_d (const void *p1, const void *p2) /* called by qsort     */
                                        /* for distance array                */
                                        /* orders on marker id & dist        */
{
   DIST_PTR pd1;
   DIST_PTR pd2;

   pd1 = (DIST_PTR const) p1;
   pd2 = (DIST_PTR const) p2;

   if (pd1->map_mark < pd2->map_mark
       || (pd1->map_mark == pd2->map_mark && pd1->dist < pd2->dist))
      return -1;

   else if (pd1->map_mark == pd2->map_mark && pd1->dist == pd2->dist)
      return 0;
   else
      return 1;
}

void stor_frq (FREQ_PTR p_f)            /* stores the number of loci         */
                                        /* for these alleles in the header   */
                                        /* and stores an array of alleles    */
{
   int   i;

   p_f->no_loci = nk_par;
   p_f->sum = 0.;

   for (i = 0; i < nx_par; ++i)
   {
      *(p_f->p_freq + i) = d_par[i];
      p_f->sum += d_par[i];
   }

   return;
}

void stor_dflt_frq (FREQ_PTR p_f)       /* stores the default allele freqs   */
{
   p_f->no_loci = 1;
   p_f->sum = 1.0;
   *(p_f->p_freq) = .5;
   *(p_f->p_freq + 1) = .5;

   return;
}

void stor_locus_data (int no, LOCUS_PTR p_l, FREQ_PTR p_f, int c, int md)
{                                       /* stores data in the locus node     */
   if (MARK_LOC == md)
   {
      if (CH == c)
         p_l->chrom_no = j_par[0];
      else
         p_l->chrom_no = 0;
   }
   else
      p_l->chrom_no = NO_CH_NO;

   p_l->no = no;
   p_l->no_alleles = nx_par;
   p_l->p_freq_node = p_f;

   return;
}

void stor_dflt_locus_data (int chro, int no, LOCUS_PTR p_l, FREQ_PTR p_f)
                                        /*stores locus node data             */
                                        /*for default allele frequencies     */
{
   p_l->chrom_no = chro;
   p_l->no = no;
   p_l->no_alleles = 2;
   p_l->p_freq_node = p_f;

   return;
}

void stor_mean_data (int no, MEANS_PTR p_node)
{
   int   i;

   p_node->no = no;
   for (i = 0; i < 3; i++)
      p_node->mean[i] = d_par[i];

   return;
}

void stor_dflt_mean_data (int no, MEANS_PTR p_node)
{
   int   i;

   p_node->no = no;
   for (i = 0; i < 3; i++)
      p_node->mean[i] = 0.;

   return;
}

void stor_vari_data (int i, VARI_PTR p_node)
{
   p_node->no = j_par[i];
   p_node->vari = d_par[0];

   return;
}

void stor_dflt_vari_data (int no, VARI_PTR p_node)
{
   p_node->no = no;
   p_node->vari = 0.;

   return;
}

void stor_mrkdat (MARKDAT_PTR p_md, int c) /* stores data in the marker    */
                                        /* data node.  Also stores the list  */
                                        /* of observed individuals and the   */
                                        /* lists of marker alleles           */
{
   int   i, j;
   char *p_n;
   int  *p_a;
   int   ix = 0;

   if (CH == c)
   {
      p_md->chrom_no = j_par[ix];
      ++ix;
   }
   else
      p_md->chrom_no = 0;

   p_md->no_obs = nm_par;
   p_md->no_marks = nk_par / (2 * nm_par);
   ++ix;

   p_n = p_md->p_obs;
   p_a = p_md->p_alls;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p_n, m_par[i], MAXNAMLEN + 1);
      p_n += MAXNAMLEN + 1;
   }
   for (j = 0; j < nk_par; ++j)
   {
      *p_a = j_par[ix++];
      ++p_a;
   }

   return;
}

void stor_mrknam (MARKNAM_PTR p_mn, int c) /* stores the chrom no, the no.  */
                                        /* of markers and the list of names   */
{
   int   i;
   char *p_n;

   if (CH == c)
      p_mn->chrom_no = j_par[0];
   else
      p_mn->chrom_no = 0;

   p_mn->no_marks = nm_par;

   p_n = p_mn->p_nams;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p_n, m_par[i], MAXNAMLEN + 1);
      p_n += MAXNAMLEN + 1;
   }

   return;
}

void stor_prbgam (PROBGAM_PTR p_pg, int cmp, int sc) /* stores data in the  */
                                        /* proband gamete node.  Also stores  */
                                        /* the array of program gametes in    */
                                        /* the proband gamete list            */
{
   int   i, ix = 0;
   PROBLIST_PTR p;

   if (CMP == cmp)
      p_pg->component = j_par[ix++];
   else
      p_pg->component = 1;

   if (SC == sc)
      p_pg->scoreset = j_par[ix++];
   else
      p_pg->scoreset = 1;

   p_pg->no_pgams = nm_par;

   p = p_pg->p_pglist;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p->nam, m_par[i], MAXNAMLEN + 1);
      p->gamete = j_par[ix++];
      ++p;
   }
   return;
}

void stor_indnam (INDNAM_PTR p_ind, int cmp) /* stores data in the named      */
                                        /* individuals node.                  */
                                        /* Also stores the names in a list    */
{
   int   i;
   char *p;

   if (CMP == cmp)
      p_ind->component = j_par[0];
   else
      p_ind->component = 1;

   p_ind->no_nams = nm_par;

   p = p_ind->p_namlist;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p, m_par[i], MAXNAMLEN + 1);
      p += MAXNAMLEN + 1;
   }
   return;
}

void stor_wind (LOCWND_PTR p_node, int c) /* stores the chromosome no.        */
                                        /* and the window size                */
{                                       /* in the locus window node           */
   int   ix = 0;

   if (CH == c)
      p_node->chrom_no = j_par[ix++];
   else
      p_node->chrom_no = 0;

   p_node->window = j_par[ix];

   return;
}

void stor_wndpat (WNDPAT_PTR p_pw, int cp, int sc) /* stores data in the    */
                                        /* window pattern node.  Also stores  */
                                        /* the window pattern labels in the   */
                                        /* list of patterns                   */
{
   int   i, ix = 0;

   if (CMP == cp)
      p_pw->component = j_par[ix++];
   else
      p_pw->component = 1;

   if (SC == sc)
      p_pw->scoreset = j_par[ix++];
   else
      p_pw->scoreset = 1;

   p_pw->no_pats = nk_par;

   for (i = 0; i < nk_par; ++i)
      *(p_pw->p_pats + i) = j_par[ix++];

   return;
}

void stor_kinshp (KINS * p_kn, int cmp, int m) /* stores data in the       */
                                        /* kinship, ibreeding or 2-locus     */
                                        /* inbreeding node.  Also stores the */
                                        /* list of pedigree members in the   */
                                        /* names list                        */
{
   int   i, ix = 0;
   char *p;

   if (CMP == cmp)
      p_kn->component = j_par[ix++];
   else
      p_kn->component = 1;

   p_kn->no_nams = nm_par;

   p = p_kn->p_nams;
   for (i = 0; i < nm_par; ++i)
   {
      if (m == 2 && i % 2)              /* checks for dup names in pair      */
         if (!(strncmp (m_par[i], m_par[i - 1], MAXNAMLEN + 1)))
         {
            err_typ = ERK_DNM;
            sprintf (errnumb, "%s", m_par[i]);
            err_txt = errnumb;
         }
      strncpy (p, m_par[i], MAXNAMLEN + 1);
      p += MAXNAMLEN + 1;
   }
   return;
}

void stor_reco_frq (int *n, double *r)  /* stores the no. of recomb freqs    */
{                                       /* and the recomb freqs              */
   int   i;

   *n = nx_par;

   for (i = 0; i < nx_par; ++i)
      r[i] = d_par[i];

   return;
}

/* ------------- copy fnc:  copies string to malloc'd memory --------------- */

char *alloc_copy (char *text)           /* first gets memory, then copies    */
{
   char *pstrg;
   int   s;

   s = strlen (text) + 1;
   pstrg = malloc (s * sizeof (char *));
   alloc_chk (WHERE, pstrg);
   strncpy (pstrg, text, s);

   return pstrg;
}

/* ---------------- reset fncs:  reinitialize data to 0's ------------------ */

void reset_sel_tr (void)                /* resets selection/nonselection     */
                                        /* of the trait and fixed effects    */
{
   int   k;

   sel_qtrait[j_par[0] - 1] = FALSE;
   for (k = 0; k < FMAX_FEF; ++k)
      seltrff[j_par[0] - 1][k] = FALSE;
   return;
}
void reset_start_ff (int m)             /* resets starting value(s) for mean */
                                        /* or fixed effects for the trait    */
{
   int   k;

   if (!m)                              /* mean                              */
   {
      nstartff[j_par[0] - 1][0] = 0;
      startff[j_par[0] - 1][0][0] = 0.0;
      return;
   }
   else
   {                                    /* fixed effect                      */
      nstartff[j_par[0] - 1][j_par[1]] = 0;
      for (k = 0; k < LEVELS; ++k)
         startff[j_par[0] - 1][j_par[1]][k] = 0.0;
      return;
   }
}
void reset_start_var (int m)            /* resets the starting value(s)      */
                                        /* for variances                     */
{
   int   j;

   for (j = 1; j <= pmx->max_trait; ++j)
      if (!m)                           /* additive                          */
         SIGSQ (j, j) = 0.0;
      else                              /* residual                          */
         TAUSQ (j, j) = 0.0;
   if (!m)
      naddvar = 0;
   else
      nresvar = 0;
   return;
}
void reset_start_cov (int m)            /* resets the starting value(s)      */
{                                       /* for covariances                   */
   int   row, col;

   for (row = 2; row <= pmx->max_trait; ++row)
      for (col = 1; col < row; ++col)
         if (!m)                        /* additive                          */
            SIGSQ (row, col) = 0.0;
         else                           /* residual                          */
            TAUSQ (row, col) = 0.0;
   if (!m)
      naddcov = 0;
   else
      nrescov = 0;
   return;
}
void reset_fit_var (int m)              /* resets the fit/nofit flags        */
{                                       /* for variances ( = 1)              */
   int   j;

   for (j = 1; j <= pmx->max_trait; ++j)
      if (!m)                           /* additive                          */
         SIG_MODL (j, j) = 1;
      else                              /* residual                          */
         TAU_MODL (j, j) = 1;
}
void reset_fit_cov (int m)              /* resets the fit/nofit flags        */
{                                       /* for covariances                   */
   int   row, col;

   for (row = 2; row <= pmx->max_trait; ++row)
      for (col = 1; col < row; ++col)
         if (!m)                        /* additive                          */
            SIG_MODL (row, col) = 1;
         else                           /* residual                          */
            TAU_MODL (row, col) = 1;
   if (!m)
      nfitadd = 0;
   else
      nfitres = 0;
   return;
}
void reset_set_eigen (void)             /* frees the eigenvalue array        */
{
   free ((void *) peigen);
   peigen = NULL;
   return;
}
void reset_set_pseudo (void)            /* frees the pseudo-prior array      */
{
   free ((void *) PseudoPrior);
   PseudoPrior = NULL;
   return;
}

/* -------- set fncs:  store input parameters into data structures --------- */

void set_sel_tr (int m)                 /* identifies the trait and fixed    */
{                                       /* effects, if any, selected         */
   int   k;

   sel_qtrait[j_par[0] - 1] = TRUE;

   if (m)
      for (k = 1; k <= nk_par; ++k)
      {
         if (seltrff[j_par[0] - 1][j_par[k]])
            err_typ = ERR_FFD;
         else
            seltrff[j_par[0] - 1][j_par[k]] = TRUE;
      }
   return;
}
void set_start_ff (int m)               /* sets the starting value(s)        */
{                                       /* for the mean or fixed effects     */
   int   k;                             /* for the trait                     */

   if (!m)                              /* mean                              */
   {
      nstartff[j_par[0] - 1][0] = 1;
      startff[j_par[0] - 1][0][0] = d_par[0];
   }
   else                                 /* fixed effect                      */
      for (k = 0; k < nx_par; ++k)
      {
         ++nstartff[j_par[0] - 1][j_par[1]];
         startff[j_par[0] - 1][j_par[1]][k] = d_par[k];
      }
   return;
}
void set_start_var (int m)              /* sets the starting value(s)        */
{                                       /* for variances                     */
   int   j;

   for (j = 1; j <= nx_par; ++j)
      if (!m)                           /* additive                          */
         SIGSQ (j, j) = d_par[j - 1];
      else                              /* residual                          */
         TAUSQ (j, j) = d_par[j - 1];
   if (!m)
      naddvar = nx_par;
   else
      nresvar = nx_par;

   return;
}
void set_start_cor (int m)              /* sets the starting value(s)        */
{                                       /* for correlations                  */
   int   j, row, col;

   for (row = 2, j = 0; j < nx_par; ++row)
      for (col = 1; col < row && j < nx_par; ++col, ++j)
         if (!m)                        /* additive                          */
            SIGSQ (row, col) = d_par[j];
         else                           /* residual                          */
            TAUSQ (row, col) = d_par[j];
   if (!m)
   {
      naddcov = nx_par;
      add_cor_flag = 1;
   }
   else
   {
      nrescov = nx_par;
      res_cor_flag = 1;
   }
   return;
}
void set_start_cov (int m)              /* sets the starting value(s)        */
{                                       /* for covariances                   */
   int   j, row, col;

   for (row = 2, j = 0; j < nx_par; ++row)
      for (col = 1; col < row && j < nx_par; ++col, ++j)
         if (!m)                        /* additive                          */
            SIGSQ (row, col) = d_par[j];
         else                           /* residual                          */
            TAUSQ (row, col) = d_par[j];
   if (!m)
   {
      naddcov = nx_par;
      add_cor_flag = 0;
   }
   else
   {
      nrescov = nx_par;
      res_cor_flag = 0;
   }
   return;
}
void set_fit_cov (int m)                /* sets the fit/no-fit flag for      */
{                                       /* covariances                       */
   int   j, row, col;

   for (row = 2, j = 0; j < nk_par; ++row)
      for (col = 1; col < row && j < nk_par; ++col, ++j)
         if (!m)                        /* additive                          */
         {
            if (j_par[j] != 0)
               SIG_MODL (row, col) = 1;
            else
               SIG_MODL (row, col) = 0;
         }
         else
         {                              /* residual                          */
            if (j_par[j] != 0)
               TAU_MODL (row, col) = 1;
            else
               TAU_MODL (row, col) = 0;
         }
   if (!m)
      nfitadd = nk_par;
   else
      nfitres = nk_par;

   return;
}
void set_set_eigen (void)               /* allocates the array for the       */
                                        /* eigenvalues and stores the        */
                                        /* starting values                   */
{
   int   k;

   peigen = malloc ((nx_par + 1) * sizeof (double));
   alloc_chk (WHERE, peigen);
   for (k = 1; k <= nx_par; ++k)
      peigen[k] = d_par[k - 1];
   n_eigen = nx_par;

   return;
}
void set_set_pseudo (void)              /* sets the pseudo-prior input flag  */
                                        /* allocates the array for the       */
                                        /* pseudo-priors and stores the      */
                                        /* values                            */
{
   int   k;

   InPseudo = Y;
   PseudoPrior = malloc ((nx_par) * sizeof (double));
   alloc_chk (WHERE, PseudoPrior);
   for (k = 0; k < nx_par; ++k)
      PseudoPrior[k] = d_par[k];
   n_pseudo = nx_par;

   return;
}
