/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 * proc_seeds() calls functions to set parameters for seeds
 */

int proc_seeds (void)
{
   if (!test_seeds ())
      if (proc_err ())
         return FALSE;

   if (!test_oseeds ())
      if (proc_err ())
         return FALSE;

   if (n_err)
      return proc_end ();

   return TRUE;
}

/* --------- test fncs:  test results of processing seed statements -------- */

int test_seeds (void)
/* determines whether seeds were given */
{
   if (GENEDROP)
   {
      if (!MarkSeedSet)                 /* tests for marker seeds set        */
      {
         MarkSeedC = CONGRUENTIAL_SEED;
         MarkSeedT = TAUSWORTHE_SEED;
         err_typ = ERD_SSD;
      }
      /*    fprintf(stdout,"\n Hexadecimal marker seeds:  %#x %#x\n\n", 
       *            MarkSeedC, MarkSeedT);  */

      if (!TraitSeedSet)                /* tests for trait seeds set         */
      {
         TraitSeedC = 0;
         TraitSeedT = 0;
      }
      else if (NTraits)
         ;
      /*       fprintf(stdout,"\n Hexadecimal trait seeds:  %#x %#x\n\n",
       *               TraitSeedC, TraitSeedT); */
   }

   if (MARKERDROP)
   {
      if (!TraitSeedSet)                /* tests for trait seeds set         */
      {
         TraitSeedC = CONGRUENTIAL_SEED;
         TraitSeedT = TAUSWORTHE_SEED;
         err_typ = ERD_STD;
      }
      /*    fprintf(stdout,"\n Hexadecimal trait seeds:  %#x %#x\n\n", 
       *            TraitSeedC, TraitSeedT);  */

      if (!MarkSeedSet)                 /* tests for marker seeds set        */
      {
         MarkSeedC = 0;
         MarkSeedT = 0;
      }
      /*    fprintf(stdout,"\n Hexadecimal marker seeds:  %#x %#x\n\n", 
       *            MarkSeedC, MarkSeedT);  */
   }

   if (LM_IBDTESTS)
   {
      if (PermutEnable)                 /*tests for permutation tests        */
      {
         if (!PermSeedSet)              /*tests for permutation seeds set    */
         {
            PermSeedC = CONGRUENTIAL_SEED;
            PermSeedT = TAUSWORTHE_SEED;
            err_typ = ERA_SSP;
         }
         else;
         /*    fprintf(stdout,"\n Hexadecimal permutation seeds:  %#x %#x\n\n", 
          *            PermSeedC, PermSeedT);  */

         if (!SamplSeedSet)             /*tests for sample seeds set         */
         {
            SamplSeedC = 0;
            SamplSeedT = 0;
         }
         else;
         /*    fprintf(stdout,"\n Hexadecimal sample seeds:  %#x %#x\n\n", 
          *            PermSeedC, PermSeedT);  */
      }
      else                              /* sets sampling seeds only          */
      {
         PermSeedC = 0;
         PermSeedT = 0;

         if (!SamplSeedSet)             /*tests for sample seeds set          */
         {
            SamplSeedC = CONGRUENTIAL_SEED;
            SamplSeedT = TAUSWORTHE_SEED;
            err_typ = ERA_SSS;
         }
         else;
         /*    fprintf(stdout,"\n Hexadecimal sample seeds:  %#x %#x\n\n", 
          *            PermSeedC, PermSeedT);  */
      }
   }
   if (IBDDROP || LM_MAP || LM_PVAL || LM_AUTO
       || LM_LODS || LM_BAYES || LM_MARKERS || LM_SCHNELL || LM_TWOQTL)
   {
      if (!SamplSeedSet)                /*tests for sampler seeds set        */
      {
         SamplSeedC = CONGRUENTIAL_SEED;
         SamplSeedT = TAUSWORTHE_SEED;
         err_typ = ERA_SSS;
      }
      /*    fprintf(stdout,"\n Hexadecimal sampler seeds:  %#x %#x\n\n", 
       *            SamplSeedC, SamplSeedT);  */
   }
   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

int test_oseeds (void)
{
   pft = get_filent (OSEED);

   if (GENEDROP || MARKERDROP)
   {
      if (pft->fname && !MarkSeedsOut && !TraitSeedsOut)
      {
         if (NLoci - NTraits)
            MarkSeedsOut = Y;
         if (NTraits)
            TraitSeedsOut = Y;
      }
      if (!pft->fname && MarkSeedsOut)
      {
         MarkSeedsOut = N;
         err_typ = ERD_OSM;
      }
      if (!pft->fname && TraitSeedsOut)
      {
         TraitSeedsOut = N;
         err_typ = ERD_OST;
      }
   }

   else if (IBDDROP || LM_MAP || LM_PVAL || LM_AUTO
            || LM_LODS || LM_BAYES || LM_MARKERS || LM_SCHNELL || LM_TWOQTL)
   {
      if (pft->fname)
         SamplSeedsOut = Y;
      else
         SamplSeedsOut = N;
   }

   else if (LM_IBDTESTS)
   {
      if (pft->fname && !SamplSeedsOut && !PermSeedsOut)
      {
         SamplSeedsOut = Y;
         PermSeedsOut = Y;
      }
      if (!pft->fname && SamplSeedsOut)
      {
         SamplSeedsOut = N;
         err_typ = ERA_USS;
      }
      if (!pft->fname && PermSeedsOut)
      {
         PermSeedsOut = N;
         err_typ = ERA_USP;
      }
   }
   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}
