/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "msample.h"

/*********************************************************/
/* This function is modified by LPT,  Nov, 2004 */
/* Calculate the fwd variables when considering multiple meiosis */
/* Rewritten by EAT, Aug, 2006, to peel each meiosis separately   */

double fwd_compute (int meio, int j, int mm, double *r, int *sx,
                    double *pen, double *prior, double *new)
/* at locus position j, given penetrances, and the priors to locus
   position j-1 (prior),
   compute the probs to locus j (new), at recombination freq r  */
{
   int   i, ii, ii1, m2, *w2;
   double temp, *tnew;

   al_int_vector (&w2, meio);
   m2 = pow2 (meio);
   al_double_vector (&tnew, m2);

   if (j == mm)
      for (ii = 0; ii < m2; ii++)
         new[ii] = pen[ii] / m2;
   else
   {
      for (ii1 = 0; ii1 < m2; ii1++)
         tnew[ii1] = prior[ii1];

      for (i = 0; i < meio; i++)
      {
         for (ii1 = 0; ii1 < meio; ii1++)
            w2[ii1] = 0;
         for (ii = 0; ii < m2; ii++)
         {
            if (w2[i] == 0)
               ii1 = ii + pow2 (meio - i - 1);
            else
               ii1 = ii - pow2 (meio - i - 1);
            new[ii] = tnew[ii] * (1.0 - r[sx[i]]) + tnew[ii1] * r[sx[i]];

            biadd (w2, meio, 1);
         }
         for (ii = 0; ii < m2; ++ii)
            tnew[ii] = new[ii];
      }
      if (pen)
         for (ii = 0; ii < m2; ii++)
            new[ii] = tnew[ii] * pen[ii];
   }

   temp = 0;
   for (i = 0; i < m2; i++)
      temp += new[i];

   for (i = 0; i < m2; i++)
      new[i] = new[i] / temp;

   free_1d ((void *) &w2);
   free_1d ((void *) &tnew);

   return temp;
}

/*********************************************************/
/* This function is modified by LPT, Nov 2004 */
/* Sample the meiosis indicators S_j when considering multiple meiosis */

void back_sample (int meio, int nloc, int j, int *right, double *r,
                  int *sx, double *left_prob, int *xx)
/* realise at locus position j, given realisation to the right, at
   recomb freq r, and cumulative penetrance to the left of left_prob
   for value 0 */
{
   double temp, *cc;
   int   i, ii, m2;

   int  *xxtemp;

   m2 = pow2 (meio);

   al_double_vector (&cc, m2);
   al_int_vector (&xxtemp, meio);

   for (i = 0; i < meio; i++)
      xx[i] = 0;
   if (j == nloc)
   {
      i = 0;
      temp = rans ();
      while ((i < m2) && (temp > left_prob[i]))
      {
         temp = temp - left_prob[i];
         i++;
         biadd (xx, meio, 1);
      }
   }
   else
   {
      temp = 0;
      for (i = 0; i < m2; i++)
      {
         cc[i] = left_prob[i];
         for (ii = 0; ii < meio; ii++)
            if (right[ii] == xx[ii])
               cc[i] *= (1 - r[sx[ii]]);
            else
               cc[i] *= r[sx[ii]];
         biadd (xx, meio, 1);
         temp += cc[i];
      }
      for (i = 0; i < m2; i++)
         cc[i] = cc[i] / temp;
      i = 0;
      temp = rans ();
      while (temp > cc[i] && i < m2)
      {
         temp = temp - cc[i];
         i++;
         biadd (xx, meio, 1);
      }
   }

   free_1d ((void *) &cc);
   free_1d ((void *) &xxtemp);

   return;
}
