/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "lsample.h"

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
void set_seq_imp (int nsmpls, int mm, int nped, int numloc, double var_err)
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
     /* Written by: AWG, April, 2001
        Modified to include linked traits, by MAJ, May, 2001
        Purpose:
        Previously S_M was initalized by assuming unlinked loci.
        Sequential imputation is being used to come up with a better
        starting value for S_M.

        Method:
        Using sequential imputation, generate S_j^* and associated weights
        w_j (i.e. just single locus peeling). Then S^* and w^* are easily
        obtained.  The imputation with maximum weight is chosen -- since
        otherwise all samples must be stored which is a BAD idea.

        Variables:
        nsmpls - number of S_M's in which to impute via sequential imputation
        nped - number of individuals in the pedigree
        numloc - number of loci (markers + trait)
        y - observed marker and trait data
        trt_ctrd - centered quantitative trait values (ym) -- obsolete EAT 8/04
        var_err - residual variance (ve)

        mm flag to allow not sampling locus 0 (mm=0) added: EAT July 2005
      */
{
   double *seq_imp_w;                   /* ptr to array of nsmpls weights */
   double weight;
   int **samp_xx0, **samp_xx1;          /* array of maternal and paternal
                                           meiosis indicators respectively */
   double mxwght = 0.0;                 /* max weight */

   int   rnloc;                         /* randomly selected locus */

   double *xecopy[2];                   /* copy of xecomo[][0][]; but malloced
                                           [2][numloc] */
   int   j, k, sx, kk;

   samp_xx0 = malloc (numloc * sizeof (int *));
   alloc_chk (WHERE, samp_xx0);
   samp_xx1 = malloc (numloc * sizeof (int *));
   alloc_chk (WHERE, samp_xx1);

   samp_xx0[0] = malloc (nped * numloc * sizeof (int));
   alloc_chk (WHERE, samp_xx0[0]);
   samp_xx1[0] = malloc (nped * numloc * sizeof (int));
   alloc_chk (WHERE, samp_xx1[0]);

   for (j = 1; j < numloc; ++j)
   {
      samp_xx0[j] = samp_xx0[j - 1] + nped;
      samp_xx1[j] = samp_xx1[j - 1] + nped;
   }

   /* and allocate array xecopy[2][] for copy of xecomo[][0][] */
   /* copy xecomo[][0][] to the array */

   for (sx = FEM; sx <= MAL; sx++)
   {
      xecopy[sx] = malloc ((numloc) * sizeof (double));
      alloc_chk (WHERE, xecopy[sx]);
      for (kk = 0; kk < numloc; kk++)
         xecopy[sx][kk] = xecomo[sx][0][kk];
   }

   set_locpeel (numloc - 1);

   seq_imp_w = malloc (numloc * sizeof (double));
   alloc_chk (WHERE, seq_imp_w);

   for (k = 0; k < nsmpls; k++)         /* draw nsmpls via seq. imp. */
   {
      /* Begin with loci on seperate linkage groups
         i.e. locus rec re-initialized */
      for (sx = FEM; sx <= MAL; sx++)
         for (kk = 0; kk < numloc; kk++)
            xecomo[sx][0][kk] = 0.5;

      /* specify map for L sampler */
      set_positions (numloc - 1);

      set_transmission (0, 0, 0);

      /* randomly select a marker or trait locus (if mm=0) 
         upon which to commence sequential imputation */
      rnloc = intran (mm, numloc - 1);

      for (kk = mm; kk < numloc; kk++)
         seq_imp_w[kk] = 0.;

      /* generate S's (as if unlinked) for randomly selected locus */
      fflush (stdout);
      seq_imp_w[rnloc] = locus_mjrpeel (var_err, 1.0, 1, 0, NULL, loci + rnloc);

      fflush (stdout);
      if (seq_imp_w[rnloc] == -DBL_MAX)
         give_up_where (WHERE, ERA_wgt, rnloc);

      /* Backwards sequential imputation */
      for (kk = rnloc - 1; kk >= mm; kk--)
      {
         for (sx = FEM; sx <= MAL; sx++)
            xecomo[sx][0][kk] = xecopy[sx][kk];
         set_positions (numloc - 1);
         fflush (stdout);
         set_transmission (loci, kk, numloc);
         seq_imp_w[kk] = locus_mjrpeel (var_err, 1.0, 1, 0, NULL, loci + kk);
         fflush (stdout);
         if (seq_imp_w[kk] == -DBL_MAX)
            give_up_where (WHERE, ERA_wgt, kk);
      }

      /* Forwards sequential imputation */
      for (kk = rnloc + 1; kk < numloc; kk++)
      {
         for (sx = FEM; sx <= MAL; sx++)
            xecomo[sx][0][kk - 1] = xecopy[sx][kk - 1];
         set_positions (numloc - 1);
         fflush (stdout);
         set_transmission (loci, kk, numloc);
         seq_imp_w[kk] = locus_mjrpeel (var_err, 1.0, 1, 0, NULL, loci + kk);
         fflush (stdout);
         if (seq_imp_w[kk] == -DBL_MAX)
            give_up_where (WHERE, ERA_wgt, kk);
      }

      /* add seq_imp_w in preparation for the calculation of the weight
         for S_M^*  */
      weight = 0.0;
      for (kk = mm; kk < numloc; kk++)
         weight += seq_imp_w[kk];

      if (k == 0 || mxwght < weight)
      {
         mxwght = weight;

         for (j = mm; j < numloc; j++)
            for (kk = 0; kk < nped; kk++)
            {
               samp_xx0[j][kk] = xx0[j][kk];
               samp_xx1[j][kk] = xx1[j][kk];
            }
      }
   }                                    /* end sequential imputation loop */

   if (seq_imp_w)
      free (seq_imp_w);

   /* assign global vars xx0 and xx1 */

   for (j = 0; j < numloc; j++)
      for (kk = 0; kk < nped; kk++)
      {
         xx0[j][kk] = samp_xx0[j][kk];
         xx1[j][kk] = samp_xx1[j][kk];
      }

   /* free structures */

   free (samp_xx0[0]);
   free (samp_xx1[0]);

   free (samp_xx0);
   free (samp_xx1);

   for (sx = FEM; sx <= MAL; sx++)
      free (xecopy[sx]);

   return;
}
