/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "stuff.h"

#include <unistd.h>                     /* For definition of gethostname */
#ifdef _WINDOWS_
#include <winsock.h>
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif

#ifdef _WINDOWS_
static int isWinNT ()
{
   OSVERSIONINFO OsVersionInfo;
   OsVersionInfo.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
   GetVersionEx (&OsVersionInfo);
   return OsVersionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT;
}

static int FileTimeToMilliseconds (FILETIME * pftime)
{
   ULARGE_INTEGER tmp;
   tmp.LowPart = pftime->dwLowDateTime;
   tmp.HighPart = pftime->dwHighDateTime;
   /* Not accurate when tmp is too large for a 32-bit integer */
   return (int) tmp.QuadPart / 10000;
}
#endif

static time_t start_tim;                /* start time for program execution */

/* ---------- print_head:  prints header at start of output ----------- */

void print_head (char *program)
{
   time (&start_tim);
   printf ("\n %s  %s", MORGVERS, program);
   printf ("               %s\n\n", ctime (&start_tim));
   fflush (stdout);
}

/* ----------- print_tail:  prints trailer at end of output -----------  */

void print_tail (void)
{
   time_t end_time;
   int   hr, minute, sec;
   double elapsd_time;

#ifdef _WINDOWS_
   WORD  wVersionRequested;
   WSADATA wsaData;

   HANDLE htHandle;
   FILETIME creationTime;
   FILETIME exitTime;
   FILETIME kernelTime;
   FILETIME userTime;
   int   err;
#else
   struct tms buffer;
   double ticks_per_sec = sysconf (_SC_CLK_TCK);
#endif /* _WINDOWS_ */
   char *hostname = malloc (MAXHOSTNAMELEN * sizeof (char));
   alloc_chk (WHERE, hostname);

#ifdef _WINDOWS_
   /* Had to initialize Winsock DLL before making any winsock calls */
   wVersionRequested = MAKEWORD (2, 0);
   err = WSAStartup (wVersionRequested, &wsaData);
   if (err == 0)
   {
      gethostname (hostname, MAXHOSTNAMELEN);
   }
   else
   {
      fprintf (stderr, "We couldn't find a usable WinSock DLL.\n");
      strcpy (hostname, "dumbwindows");
   }
   /* clean up to allow others use the DLL */
   WSACleanup ();
#else
   gethostname (hostname, MAXHOSTNAMELEN);
#endif /* _WINDOWS_ */

   printf ("\n     hostname:  %s\n\n", hostname);
   free (hostname);

   hr = minute = sec = 0;
   time (&end_time);
   elapsd_time = difftime (end_time, start_tim);

   hr = elapsd_time / 3600;

   if (elapsd_time >= 3600 * hr)
      elapsd_time -= 3600 * hr;

   minute = elapsd_time / 60;

   if (elapsd_time >= 60 * minute)
      elapsd_time -= 60 * minute;

   sec = elapsd_time;
#ifdef _WINDOWS_
   if (isWinNT ())
   {
      /* GetProcessTimes is not implemented for Win 95/98/Me */
      htHandle = GetCurrentProcess ();
      err = GetProcessTimes (htHandle,
                             &creationTime, &exitTime, &kernelTime, &userTime);
      if (err == FALSE)
      {
         fprintf (stderr, "Error in GetProcessTimes: 0x%lx\n",
                  (int) GetLastError ());
         fflush (stderr);
      }
      printf ("  system time:  %.3f sec\n",
              (double) FileTimeToMilliseconds (&kernelTime) / 1000);
      printf ("    user time:  %.3f sec\n",
              (double) FileTimeToMilliseconds (&userTime) / 1000);
   }
#else
   times (&buffer);
   printf ("  system time:  %.3f sec\n", buffer.tms_stime / ticks_per_sec);
   printf ("    user time:  %.3f sec\n", buffer.tms_utime / ticks_per_sec);
#endif /* _WINDOWS_ */
   printf (" elapsed time:  %d hr %d min %d sec\n\n", hr, minute, sec);
   fflush (stdout);

   fflush (stderr);
}

/* vim: sw=3
 */
