/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 NAME
      print_S, print_S_marker, print_mc_eval

 SYNOPSIS

      #include "twoqtl.h"
      void print_S (char **S_0, char **S_1, int n_locus, int comp_no);
      void print_S_marker (int comp_no);
      void print_mc_eval (mc_eval_t out);

 DESCRIPTION 
      print intermediate results in lm_twoqtl
      (just for testing and debugging, not for the ordinary user)

 YJS 07/02/2005     
*/
#include "twoqtl.h"

static int *size_c = NULL, *ibase = NULL;

void print_S (char **S_1, char **S_0, int n_locus, int comp_no)
{
   int   j_stop, locus, j;
   char *S_1_locus, *S_0_locus;

   if (size_c == NULL)
      size_c = get_size_c ();

   fprintf (stderr, "\n S_trait of Component %2d", comp_no);
   j_stop = size_c[comp_no];

   for (locus = 0; locus < n_locus; ++locus)
   {
      S_1_locus = S_1[locus];
      S_0_locus = S_0[locus];

      fprintf (stderr, "\n %2d", locus);
      for (j = 0; j < j_stop; ++j)
         fprintf (stderr, " %2d%2d", (int) S_1_locus[j], (int) S_0_locus[j]);
   }
   fprintf (stderr, "\n");
   return;
}

void print_S_marker (int comp_no)
{
   int   j_start, j_stop, locus, j;
   int  *S_1_locus, *S_0_locus;

   if (size_c == NULL)
      size_c = get_size_c ();
   if (ibase == NULL)
      ibase = get_ibase ();

   fprintf (stderr, "\n S_marker of Component %2d", comp_no);
   j_start = ibase[comp_no];
   j_stop = j_start + size_c[comp_no];

   for (locus = 0; locus < NMarks; ++locus)
   {
      /* note: array index of xx1, xx0 begins at 1 (for markers) */
      S_1_locus = xx1[locus + 1];
      S_0_locus = xx0[locus + 1];

      fprintf (stderr, "\n %2d", locus);
      for (j = j_start; j < j_stop; ++j)
         fprintf (stderr, " %2d%2d", S_1_locus[j], S_0_locus[j]);
   }
   fprintf (stderr, "\n");

   return;
}

void print_mc_eval (mc_eval_t out)
{
   int   n_loci_1, n_loci_2, i, j;
   double *foo_alt, *max_alt;

   n_loci_1 = out.n_loci_1;
   n_loci_2 = out.n_loci_2;

   fprintf (stderr, "print mc_eval (n_loci_1 = %2d, n_loci_2 = %2d)\n",
            n_loci_1, n_loci_2);

   for (i = 0; i < n_loci_1; ++i)
   {
      foo_alt = out.foo_alt[i];
      max_alt = out.max_alt[i];

      fprintf (stderr, "%2d ", i);
      for (j = 0; j < n_loci_2; ++j)
         fprintf (stderr, "%e %e ", foo_alt[j], max_alt[j]);
      fprintf (stderr, "\n");
   }
   fprintf (stderr, "nl %e %e \n", out.foo_null, out.max_null);

   return;
}
